from arcor2_arserver.lock import exceptions, notifications, structures
from arcor2_arserver.lock.lock import Lock

__all__ = [Lock.__name__, exceptions.__name__, notifications.__name__, structures.__name__]



import asyncio
import inspect
from typing import Any

import arcor2.helpers as hlp
from arcor2.cached import CachedScene
from arcor2.clients import aio_asset as asset
from arcor2.data import common
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import MultiArmRobot, Robot
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver import objects_actions as osa
from arcor2_arserver.object_types.data import ObjectTypeData
from arcor2_arserver.object_types.source import function_implemented
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data.robot import RobotMeta


class RobotPoseException(Arcor2Exception):
    pass


class SingleArmRobotException(Arcor2Exception):
    pass


def prepare_args(robot_inst: Robot, args: list[Any], arm_id: None | str) -> list[Any]:

    if isinstance(robot_inst, MultiArmRobot):
        args.append(arm_id)
    elif arm_id:
        raise SingleArmRobotException("Single arm robot.")

    return args


async def get_arms(robot_inst: Robot) -> set[str]:
    """
    :param robot_inst:
    :return: IDs of existing arms.
    """

    if not isinstance(robot_inst, MultiArmRobot):
        raise SingleArmRobotException(f"{robot_inst.__class__.__name__} has only one arm.")

    return await hlp.run_in_executor(robot_inst.get_arm_ids)


async def get_end_effectors(robot_inst: Robot, arm_id: None | str) -> set[str]:
    """
    :param robot_inst:
    :return: IDs of existing end effectors.
    """

    return await hlp.run_in_executor(robot_inst.get_end_effectors_ids, *prepare_args(robot_inst, [], arm_id))


async def get_grippers(robot_inst: Robot, arm_id: None | str) -> set[str]:
    """
    :param robot_inst:
    :return: IDs of existing grippers.
    """

    return await hlp.run_in_executor(robot_inst.grippers, *prepare_args(robot_inst, [], arm_id))


async def get_suctions(robot_inst: Robot, arm_id: None | str) -> set[str]:
    """
    :param robot_inst:
    :return: IDs of existing suctions.
    """

    return await hlp.run_in_executor(robot_inst.suctions, *prepare_args(robot_inst, [], arm_id))


async def get_pose_and_joints(
    robot_inst: Robot, end_effector: str, arm_id: None | str
) -> tuple[common.Pose, list[common.Joint]]:

    return await asyncio.gather(
        hlp.run_in_executor(robot_inst.get_end_effector_pose, *prepare_args(robot_inst, [end_effector], arm_id)),
        hlp.run_in_executor(robot_inst.robot_joints, *prepare_args(robot_inst, [False], arm_id)),
    )


async def get_end_effector_pose(robot_inst: Robot, end_effector: str, arm_id: None | str) -> common.Pose:
    """
    :param robot_inst:
    :param end_effector:
    :return: Global pose
    """

    return await hlp.run_in_executor(
        robot_inst.get_end_effector_pose, *prepare_args(robot_inst, [end_effector], arm_id)
    )


async def get_robot_joints(robot_inst: Robot, arm_id: None | str, include_gripper: bool = False) -> list[common.Joint]:
    """
    :param robot_inst:
    :return: List of joints
    """

    return await hlp.run_in_executor(robot_inst.robot_joints, *prepare_args(robot_inst, [include_gripper], arm_id))


def _feature(type_def: type[Robot], method_name: str, base_class: type[Robot]) -> bool:

    assert glob.OBJECT_TYPES

    method = getattr(type_def, method_name)
    where_it_is_defined = glob.OBJECT_TYPES.get(method.__qualname__.split(".")[0], None)

    if where_it_is_defined is None:
        raise Arcor2Exception(f"Can't get origin for {type_def.__name__}/{method_name}.")

    if where_it_is_defined.type_def is None or where_it_is_defined.ast is None:
        raise Arcor2Exception(
            f"Origin {where_it_is_defined.meta.type} for {type_def.__name__}/{method_name} is disabled."
        )

    logger.debug(
        f"Processing {type_def.__name__}/{method_name} "
        f"(defined in {where_it_is_defined.type_def.__name__}), with base class {base_class.__name__}."
    )

    if where_it_is_defined.type_def is base_class or where_it_is_defined.meta.disabled:
        # all of the "feature" methods are abstract in the base class and have to be implemented for the concrete robot
        return False

    if not function_implemented(where_it_is_defined.ast, method_name):
        logger.debug(f"{type_def.__name__}/{method_name} not implemented.")
        return False

    sign = inspect.signature(getattr(where_it_is_defined.type_def, method_name))
    if not (res := inspect.signature(getattr(base_class, method_name)) == sign):
        logger.debug(f"{type_def.__name__}/{method_name} has invalid signature.")
    return res


async def get_robot_meta(obj_type: ObjectTypeData) -> None:

    if obj_type.meta.disabled:
        raise Arcor2Exception("Disabled object type.")

    assert obj_type.type_def is not None

    if not issubclass(obj_type.type_def, Robot):
        raise Arcor2Exception("Not a robot.")

    obj_type.robot_meta = RobotMeta(
        obj_type.meta.type, obj_type.type_def.robot_type, issubclass(obj_type.type_def, MultiArmRobot)
    )

    # TODO fix mypy issue 'Can only assign concrete classes to a variable of type "Type[Robot]"'
    base_class: type[Robot] = MultiArmRobot if obj_type.robot_meta.multi_arm else Robot  # type: ignore

    # TODO automate this somehow
    obj_type.robot_meta.features.move_to_pose = _feature(obj_type.type_def, Robot.move_to_pose.__name__, base_class)
    obj_type.robot_meta.features.move_to_joints = _feature(obj_type.type_def, Robot.move_to_joints.__name__, base_class)
    obj_type.robot_meta.features.stop = _feature(obj_type.type_def, Robot.stop.__name__, base_class)
    obj_type.robot_meta.features.inverse_kinematics = _feature(
        obj_type.type_def, Robot.inverse_kinematics.__name__, base_class
    )
    obj_type.robot_meta.features.forward_kinematics = _feature(
        obj_type.type_def, Robot.forward_kinematics.__name__, base_class
    )
    obj_type.robot_meta.features.hand_teaching = _feature(
        obj_type.type_def, Robot.set_hand_teaching_mode.__name__, base_class
    )

    if urdf_name := obj_type.type_def.urdf_package_name:

        if not await asset.asset_exists(urdf_name):
            logger.error(f"URDF package {urdf_name} for {obj_type.meta.type} does not exist.")
        else:
            obj_type.robot_meta.urdf_package_filename = urdf_name
            # TODO check if URDF is valid?

    logger.debug(obj_type.robot_meta)


async def stop(robot_inst: Robot) -> None:

    if not robot_inst.move_in_progress:
        raise Arcor2Exception("Robot is not moving.")

    await hlp.run_in_executor(robot_inst.stop)


async def ik(
    robot_inst: Robot,
    end_effector_id: str,
    arm_id: None | str,
    pose: common.Pose,
    start_joints: None | list[common.Joint] = None,
    avoid_collisions: bool = True,
) -> list[common.Joint]:

    return await hlp.run_in_executor(
        robot_inst.inverse_kinematics,
        *prepare_args(robot_inst, [end_effector_id, pose, start_joints, avoid_collisions], arm_id),
    )


async def fk(robot_inst: Robot, end_effector_id: str, arm_id: None | str, joints: list[common.Joint]) -> common.Pose:

    return await hlp.run_in_executor(
        robot_inst.forward_kinematics, *prepare_args(robot_inst, [end_effector_id, joints], arm_id)
    )


async def check_reachability(
    scene: CachedScene,
    robot_inst: Robot,
    end_effector_id: str,
    arm_id: None | str,
    pose: common.Pose,
    safe: bool = True,
) -> None:

    otd = osa.get_obj_type_data(scene, robot_inst.id)
    if otd.robot_meta and otd.robot_meta.features.inverse_kinematics:
        try:
            await ik(robot_inst, end_effector_id, arm_id, pose, avoid_collisions=safe)
        except Robot.KinematicsException:
            raise Arcor2Exception("Unreachable pose.")
        except Arcor2Exception as e:
            logger.exception("Failed to check reachability.")
            raise Arcor2Exception("Error on getting IK.") from e


async def check_robot_before_move(robot_inst: Robot) -> None:

    robot_inst.check_if_ready_to_move()

    if glob.RUNNING_ACTION:

        assert glob.LOCK.project
        action = glob.LOCK.project.action(glob.RUNNING_ACTION)
        obj_id_str, _ = action.parse_type()

        if robot_inst.id == obj_id_str:
            raise Arcor2Exception("Robot is executing action.")


async def _move_to_pose(
    robot_inst: Robot,
    end_effector_id: str,
    arm_id: None | str,
    pose: common.Pose,
    speed: float,
    safe: bool,
    linear: bool,
) -> None:
    # TODO newly connected interface should be notified somehow (general solution for such cases would be great!)

    try:
        await hlp.run_in_executor(
            robot_inst.move_to_pose, *prepare_args(robot_inst, [end_effector_id, pose, speed, safe, linear], arm_id)
        )
    except Arcor2Exception as e:
        logger.error(f"Robot movement failed with: {str(e)}")
        raise


async def move_to_pose(
    robot_inst: Robot,
    end_effector_id: str,
    arm_id: None | str,
    pose: common.Pose,
    speed: float,
    safe: bool,
    linear: bool,
    lock_owner: None | str = None,
) -> None:

    try:
        Data = sevts.r.RobotMoveToPose.Data

        await notif.broadcast_event(
            sevts.r.RobotMoveToPose(
                Data(Data.MoveEventType.START, robot_inst.id, end_effector_id, pose, safe, linear, arm_id=arm_id)
            )
        )

        try:
            await _move_to_pose(robot_inst, end_effector_id, arm_id, pose, speed, safe, linear)

        except Arcor2Exception as e:
            await notif.broadcast_event(
                sevts.r.RobotMoveToPose(
                    Data(Data.MoveEventType.FAILED, robot_inst.id, end_effector_id, pose, safe, linear, str(e), arm_id)
                )
            )
            return

        await notif.broadcast_event(
            sevts.r.RobotMoveToPose(
                Data(Data.MoveEventType.END, robot_inst.id, end_effector_id, pose, safe, linear, arm_id=arm_id)
            )
        )
    finally:
        if lock_owner:
            await glob.LOCK.write_unlock(robot_inst.id, lock_owner)


async def move_to_ap_orientation(
    robot_inst: Robot,
    end_effector_id: str,
    arm_id: None | str,
    pose: common.Pose,
    speed: float,
    orientation_id: str,
    safe: bool,
    linear: bool,
) -> None:

    Data = sevts.r.RobotMoveToActionPointOrientation.Data

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointOrientation(
            Data(Data.MoveEventType.START, robot_inst.id, end_effector_id, orientation_id, safe, linear, arm_id=arm_id)
        )
    )

    try:
        await _move_to_pose(robot_inst, end_effector_id, arm_id, pose, speed, safe, linear)

    except Arcor2Exception as e:
        await notif.broadcast_event(
            sevts.r.RobotMoveToActionPointOrientation(
                Data(
                    Data.MoveEventType.FAILED,
                    robot_inst.id,
                    end_effector_id,
                    orientation_id,
                    safe,
                    linear,
                    str(e),
                    arm_id,
                )
            )
        )
        return

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointOrientation(
            Data(Data.MoveEventType.END, robot_inst.id, end_effector_id, orientation_id, safe, linear, arm_id=arm_id)
        )
    )


async def _move_to_joints(
    robot_inst: Robot, joints: list[common.Joint], speed: float, safe: bool, arm_id: None | str
) -> None:

    # TODO newly connected interface should be notified somehow (general solution for such cases would be great!)

    try:
        await hlp.run_in_executor(robot_inst.move_to_joints, *prepare_args(robot_inst, [joints, speed, safe], arm_id))
    except Arcor2Exception as e:
        logger.error(f"Robot movement failed with: {str(e)}")
        raise


async def move_to_joints(
    robot_inst: Robot,
    joints: list[common.Joint],
    speed: float,
    safe: bool,
    arm_id: None | str,
    lock_owner: None | str = None,
) -> None:

    try:
        Data = sevts.r.RobotMoveToJoints.Data

        await notif.broadcast_event(
            sevts.r.RobotMoveToJoints(Data(Data.MoveEventType.START, robot_inst.id, joints, safe, arm_id=arm_id))
        )

        try:

            await _move_to_joints(robot_inst, joints, speed, safe, arm_id)

        except Arcor2Exception as e:

            await notif.broadcast_event(
                sevts.r.RobotMoveToJoints(
                    Data(Data.MoveEventType.FAILED, robot_inst.id, joints, safe, str(e), arm_id=arm_id)
                )
            )

            return

        await notif.broadcast_event(
            sevts.r.RobotMoveToJoints(Data(Data.MoveEventType.END, robot_inst.id, joints, safe, arm_id=arm_id))
        )
    finally:
        if lock_owner:
            await glob.LOCK.write_unlock(robot_inst.id, lock_owner)


async def move_to_ap_joints(
    robot_inst: Robot,
    joints: list[common.Joint],
    speed: float,
    joints_id: str,
    safe: bool,
    arm_id: None | str,
) -> None:

    Data = sevts.r.RobotMoveToActionPointJoints.Data

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointJoints(
            Data(Data.MoveEventType.START, robot_inst.id, joints_id, safe, arm_id=arm_id)
        )
    )

    try:

        await _move_to_joints(robot_inst, joints, speed, safe, arm_id)

    except Arcor2Exception as e:

        await notif.broadcast_event(
            sevts.r.RobotMoveToActionPointJoints(
                Data(Data.MoveEventType.FAILED, robot_inst.id, joints_id, safe, str(e), arm_id)
            )
        )

        return

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointJoints(
            Data(Data.MoveEventType.END, robot_inst.id, joints_id, safe, arm_id=arm_id)
        )
    )


async def check_eef_arm(robot_inst: Robot, arm_id: None | str, eef_id: None | str = None) -> None:

    if isinstance(robot_inst, MultiArmRobot):
        if not arm_id:
            raise Arcor2Exception("Arm has to be specified.")

        if eef_id is not None and eef_id not in await hlp.run_in_executor(robot_inst.get_end_effectors_ids, arm_id):
            raise Arcor2Exception("Unknown end effector.")

        return

    if arm_id:
        raise Arcor2Exception("Arm should not be specified.")

    if eef_id is not None and eef_id not in await hlp.run_in_executor(robot_inst.get_end_effectors_ids):
        raise Arcor2Exception("Unknown end effector.")



import os
from collections import defaultdict
from typing import Any

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.data import events
from arcor2.object_types.abstract import Generic
from arcor2_arserver.lock import Lock
from arcor2_arserver.object_types.data import ObjectTypeDict, ObjTypeDict
from arcor2_arserver.user import Users
from arcor2_arserver_data.events.common import ShowMainScreen

VERBOSE: bool = False

PORT: int = int(os.getenv("ARCOR2_ARSERVER_PORT", 6789))

MAIN_SCREEN: None | ShowMainScreen.Data = ShowMainScreen.Data(ShowMainScreen.Data.WhatEnum.ScenesList)

OBJECT_TYPES: ObjectTypeDict = ObjTypeDict()

SCENE_OBJECT_INSTANCES: dict[str, Generic] = {}

RUNNING_ACTION: None | str = None  # ID of an action that is being executed during project editing
RUNNING_ACTION_PARAMS: None | list[Any] = None

PACKAGE_STATE = events.PackageState.Data()
PACKAGE_INFO: None | events.PackageInfo.Data = None

# there might be some long-running action being executed when ui connects, so let's them know
ACTION_STATE_BEFORE: None | events.ActionStateBefore.Data = None

TEMPORARY_PACKAGE: bool = False

RegisteredUiDict = defaultdict[str, set[WsClient]]

ROBOT_JOINTS_REGISTERED_UIS: RegisteredUiDict = defaultdict(lambda: set())  # robot, UIs
ROBOT_EEF_REGISTERED_UIS: RegisteredUiDict = defaultdict(lambda: set())  # robot, UIs

OBJECTS_WITH_UPDATED_POSE: set[str] = set()

LOCK: Lock = Lock(OBJECT_TYPES)

USERS: Users = Users()

PREV_RESULTS: dict[str, tuple[Any] | Any] = {}


def remove_prev_result(action_id: str) -> None:

    try:
        del PREV_RESULTS[action_id]
    except KeyError:
        pass



from websockets.server import WebSocketServerProtocol as WsClient

from arcor2_arserver import globals as glob
from arcor2_arserver.lock.exceptions import CannotLock
from arcor2_arserver_data import rpc as srpc


async def write_lock_cb(req: srpc.lock.WriteLock.Request, ui: WsClient) -> None:

    if not await glob.LOCK.write_lock(req.args.object_id, glob.USERS.user_name(ui), req.args.lock_tree, notify=True):
        raise CannotLock(glob.LOCK.ErrMessages.LOCK_FAIL.value)


async def write_unlock_cb(req: srpc.lock.WriteUnlock.Request, ui: WsClient) -> None:

    await glob.LOCK.write_unlock(req.args.object_id, glob.USERS.user_name(ui), notify=True)


async def read_lock_cb(req: srpc.lock.ReadLock.Request, ui: WsClient) -> None:

    # TODO currently unused, maybe delete?
    if not await glob.LOCK.read_lock(req.args.object_id, glob.USERS.user_name(ui)):
        raise CannotLock(glob.LOCK.ErrMessages.LOCK_FAIL.value)


async def read_unlock_cb(req: srpc.lock.ReadUnlock.Request, ui: WsClient) -> None:

    # TODO currently unused, maybe delete?
    await glob.LOCK.read_unlock(req.args.object_id, glob.USERS.user_name(ui))


async def update_lock_cb(req: srpc.lock.UpdateLock.Request, ui: WsClient) -> None:

    await glob.LOCK.update_lock(req.args.object_id, glob.USERS.user_name(ui), req.args.new_type)



from ast import AST
from dataclasses import dataclass, field

from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Generic
from arcor2_arserver_data.objects import ObjectAction, ObjectTypeMeta
from arcor2_arserver_data.robot import RobotMeta


@dataclass
class ObjectTypeData:

    meta: ObjectTypeMeta
    type_def: None | type[Generic] = None
    actions: dict[str, ObjectAction] = field(default_factory=dict)
    ast: None | AST = None
    robot_meta: None | RobotMeta = None

    def __post_init__(self) -> None:
        if not self.meta.disabled:
            assert self.type_def is not None
            assert self.ast is not None


class ObjTypeDict(dict):
    def __getitem__(self, object_type_id: str) -> ObjectTypeData:
        try:
            return super().__getitem__(object_type_id)
        except KeyError:
            raise Arcor2Exception(f"Unknown ObjectType: {object_type_id}.")


ObjectTypeDict = dict[str, ObjectTypeData]



from arcor2 import helpers as hlp
from arcor2.cached import CachedProject, CachedScene
from arcor2.data.common import (
    Action,
    ActionParameter,
    LogicItem,
    Parameter,
    ProjectFunction,
    ProjectParameter,
    SceneObject,
)
from arcor2.exceptions import Arcor2Exception
from arcor2.logic import LogicContainer
from arcor2.parameter_plugins import ParameterPluginException
from arcor2.parameter_plugins.utils import known_parameter_types, plugin_from_type_name
from arcor2_arserver.object_types.data import ObjectTypeData, ObjectTypeDict
from arcor2_arserver.objects_actions import get_types_dict
from arcor2_arserver_data.objects import ObjectAction

# TODO refactor the module somewhere, so it can be also used within arcor2_build?


def find_object_action(obj_types: ObjectTypeDict, scene: CachedScene, action: Action) -> ObjectAction:

    obj_id, action_type = action.parse_type()
    obj = scene.object(obj_id)
    obj_type = obj_types[obj.type]

    try:
        act = obj_type.actions[action_type]
    except KeyError:
        raise Arcor2Exception("Unknown type of action.")

    if act.disabled:
        raise Arcor2Exception("Action is disabled.")

    return act


# TODO refactor into arcor2.logic?
def check_logic_item(
    obj_types: ObjectTypeDict, scene: CachedScene, parent: LogicContainer, logic_item: LogicItem
) -> None:
    """Checks if newly added/updated ProjectLogicItem is ok.

    :param parent:
    :param logic_item:
    :return:
    """

    action_ids = parent.action_ids()

    if logic_item.start == LogicItem.START and logic_item.end == LogicItem.END:
        raise Arcor2Exception("This does not make sense.")

    if logic_item.start != LogicItem.START:

        start_action_id, start_flow = logic_item.parse_start()

        if start_action_id == logic_item.end:
            raise Arcor2Exception("Start and end can't be the same.")

        if start_action_id not in action_ids:
            raise Arcor2Exception("Logic item has unknown start.")

        if start_flow != "default":  # TODO enum
            raise Arcor2Exception("Only flow 'default' is supported so far.'")

    if logic_item.end != LogicItem.END:

        if logic_item.end not in action_ids:
            raise Arcor2Exception("Logic item has unknown end.")

    if logic_item.condition is not None:

        what = logic_item.condition.parse_what()
        action = parent.action(what.action_id)  # action that produced the result which we depend on here
        flow = action.flow(what.flow_name)
        try:
            flow.outputs[what.output_index]
        except IndexError:
            raise Arcor2Exception(f"Flow {flow.type} does not have output with index {what.output_index}.")

        action_meta = find_object_action(obj_types, scene, action)

        try:
            return_type = action_meta.returns[what.output_index]
        except IndexError:
            raise Arcor2Exception(f"Invalid output index {what.output_index} for action {action_meta.name}.")

        return_type_plugin = plugin_from_type_name(return_type)

        if not return_type_plugin.COUNTABLE:
            raise Arcor2Exception(f"Output of type {return_type} can't be branched.")

        # TODO for now there is only support for bool
        if return_type_plugin.type() != bool:
            raise Arcor2Exception("Unsupported condition type.")

        # check that condition value is ok, actual value is not interesting
        # TODO perform this check using plugin
        from arcor2 import json

        if not isinstance(json.loads(logic_item.condition.value), bool):
            raise Arcor2Exception("Invalid condition value.")

    for existing_item in parent.logic:

        if existing_item.id == logic_item.id:  # item is updated
            continue

        if logic_item.start == logic_item.START and existing_item.start == logic_item.START:
            raise Arcor2Exception("START already defined.")

        if logic_item.start == existing_item.start:

            if None in (logic_item.condition, existing_item.condition):
                raise Arcor2Exception("Two junctions has the same start action without condition.")

            # when there are more logical connections from A to B, their condition values must be different
            if logic_item.condition == existing_item.condition:
                raise Arcor2Exception("Two junctions with the same start should have different conditions.")

        if logic_item.end == existing_item.end:
            if logic_item.start == existing_item.start:
                raise Arcor2Exception("Junctions can't have the same start and end.")


def check_parameter(parameter: Parameter) -> None:

    # TODO check using (some) plugin
    from arcor2 import json

    val = json.loads(parameter.value)

    # however, analysis in get_dataclass_params() can handle also (nested) dataclasses, etc.
    if not isinstance(val, (int, float, str, bool)):
        raise Arcor2Exception("Only basic types are supported so far.")


def check_object_parameters(obj_type: ObjectTypeData, parameters: list[Parameter]) -> None:

    if {s.name for s in obj_type.meta.settings if s.default_value is None} > {s.name for s in parameters}:
        raise Arcor2Exception("Some required parameter is missing.")

    param_dict = obj_type.meta.parameters_dict()

    for param in parameters:

        if param_dict[param.name].type != param.type:
            raise Arcor2Exception(f"Type mismatch for parameter {param}.")

        check_parameter(param)


def check_project_parameter(proj: CachedProject, parameter: ProjectParameter) -> None:

    hlp.is_valid_identifier(parameter.name)

    for pparam in proj.parameters:

        if parameter.id == pparam.id:
            continue

        if parameter.name == pparam.name:
            raise Arcor2Exception(f"Project parameter name {parameter.name} is duplicate.")

    check_parameter(parameter)


def check_object(obj_types: ObjectTypeDict, scene: CachedScene, obj: SceneObject, new_one: bool = False) -> None:
    """Checks if object can be added into the scene."""

    if obj.type not in obj_types:
        raise Arcor2Exception(f"Unknown ObjectType {obj.type}.")

    obj_type = obj_types[obj.type]

    if obj_type.meta.disabled:
        raise Arcor2Exception(f"ObjectType {obj.type} is disabled. {obj_type.meta.problem}")

    check_object_parameters(obj_type, obj.parameters)

    # TODO check whether object needs parent and if so, if the parent is in scene and parent_id is set
    if obj_type.meta.needs_parent_type:
        pass

    if obj_type.meta.has_pose and obj.pose is None:
        raise Arcor2Exception(f"Object {obj.name} requires pose.")

    if not obj_type.meta.has_pose and obj.pose is not None:
        raise Arcor2Exception(f"Object {obj.name} should not have pose.")

    if obj_type.meta.abstract:
        raise Arcor2Exception(f"ObjectType {obj.type} is abstract.")

    if new_one:

        if obj.id in scene.object_ids:
            raise Arcor2Exception(f"Object {obj.name} has duplicate id.")

        if obj.name in scene.object_names():
            raise Arcor2Exception(f"Object name {obj.name} is duplicate.")

    hlp.is_valid_identifier(obj.name)


def check_action_params(
    obj_types: ObjectTypeDict, scene: CachedScene, project: CachedProject, action: Action, object_action: ObjectAction
) -> None:

    _, action_type = action.parse_type()

    assert action_type == object_action.name

    if len(object_action.parameters) != len(action.parameters):
        raise Arcor2Exception("Unexpected number of parameters.")

    for req_param in object_action.parameters:

        param = action.parameter(req_param.name)

        if param.type == ActionParameter.TypeEnum.PROJECT_PARAMETER:

            pparam = project.parameter(param.str_from_value())

            param_meta = object_action.parameter(param.name)
            if param_meta.type != pparam.type:
                raise Arcor2Exception("Action parameter type does not match project parameter type.")

        elif param.type == ActionParameter.TypeEnum.LINK:

            parsed_link = param.parse_link()

            if parsed_link.action_id == action.id:
                raise Arcor2Exception("Can't use own result as a parameter.")

            parent_action = project.action(parsed_link.action_id)
            source_action_pt = parent_action.parse_type()

            parent_action_meta = obj_types[scene.object(source_action_pt.obj_id).type].actions[
                source_action_pt.action_type
            ]

            if len(parent_action.flow(parsed_link.flow_name).outputs) != len(parent_action_meta.returns):
                raise Arcor2Exception("Source action does not have outputs specified.")

            param_meta = object_action.parameter(param.name)

            try:
                if param_meta.type != parent_action_meta.returns[parsed_link.output_index]:
                    raise Arcor2Exception("Param type does not match action output type.")
            except IndexError:
                raise Arcor2Exception(
                    f"Index {parsed_link.output_index} is invalid for action {object_action.name},"
                    f" which returns {len(object_action.returns)} values."
                )

        else:

            if param.type not in known_parameter_types():
                raise Arcor2Exception(f"Parameter {param.name} of action {action.name} has unknown type: {param.type}.")

            try:
                plugin_from_type_name(param.type).parameter_value(
                    get_types_dict(), scene, project, action.id, param.name
                )
            except ParameterPluginException as e:
                raise Arcor2Exception(f"Parameter {param.name} of action {action.name} has invalid value. {str(e)}")


def check_flows(parent: CachedProject | ProjectFunction, action: Action, action_meta: ObjectAction) -> None:
    """Raises exception if there is something wrong with flow(s).

    :param parent:
    :param action:
    :param action_meta:
    :return:
    """

    flow = action.flow()  # searches default flow (just this flow is supported so far)

    # it is ok to not specify any output (if the values are not going to be used anywhere)
    # return value(s) won't be stored in variable(s)
    if not flow.outputs:
        return

    # otherwise, all return values have to be stored in variables
    if len(flow.outputs) != len(action_meta.returns):
        raise Arcor2Exception("Number of the flow outputs does not match the number of action outputs.")

    for output in flow.outputs:
        hlp.is_valid_identifier(output)

    outputs: set[str] = set()

    for act in parent.actions:
        for fl in act.flows:
            for output in fl.outputs:
                if output in outputs:
                    raise Arcor2Exception(f"Output '{output}' is not unique.")


def scene_problems(obj_types: ObjectTypeDict, scene: CachedScene) -> list[str]:

    problems: list[str] = []

    for scene_obj in scene.objects:
        try:
            check_object(obj_types, scene, scene_obj)
        except Arcor2Exception as e:
            problems.append(str(e))

    return problems


def check_ap_parent(scene: CachedScene, proj: CachedProject, parent: None | str) -> None:

    if not parent:
        return

    if parent in scene.object_ids:
        if scene.object(parent).pose is None:
            raise Arcor2Exception("AP can't have object without pose as parent.")
    elif parent not in proj.action_points_ids:
        raise Arcor2Exception("AP has invalid parent ID (not an object or another AP).")


def project_problems(obj_types: ObjectTypeDict, scene: CachedScene, project: CachedProject) -> list[str]:

    if project.scene_id != scene.id:
        return ["Project/scene mismatch."]

    problems: list[str] = scene_problems(obj_types, scene)

    for proj_param in project.parameters:
        try:
            check_project_parameter(project, proj_param)
        except Arcor2Exception as e:
            problems.append(str(e))

    for ap in project.action_points:

        try:
            check_ap_parent(scene, project, ap.parent)
        except Arcor2Exception:
            problems.append(f"Action point {ap.name} has invalid parent: {ap.parent}.")

        for joints in project.ap_joints(ap.id):
            if joints.robot_id not in scene.object_ids:
                problems.append(
                    f"Action point {ap.name} has joints ({joints.name}) for an unknown robot: {joints.robot_id}."
                )

        for action in project.actions:

            # check if objects have used actions
            obj_id, action_type = action.parse_type()

            if obj_id not in scene.object_ids:
                problems.append(f"Object ID {obj_id} which action is used in {action.name} does not exist in scene.")
                continue

            scene_obj = scene.object(obj_id)
            if action_type not in obj_types[scene_obj.type].actions:
                problems.append(
                    f"ObjectType {scene_obj.type} does not have action {action_type} used in {action.name}."
                )
                continue

            action_meta = obj_types[scene_obj.type].actions[action_type]

            try:
                check_action_params(obj_types, scene, project, action, action_meta)
            except Arcor2Exception as e:
                problems.append(str(e))

            try:
                check_flows(project, action, action_meta)
            except Arcor2Exception as e:
                problems.append(str(e))

    return problems



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
from typing import AsyncGenerator, Iterable

from arcor2 import env
from arcor2.cached import CachedProjectException, UpdateableCachedProject, UpdateableCachedScene
from arcor2.data import common as cmn
from arcor2.exceptions import Arcor2Exception
from arcor2_arserver import logger
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.lock.common import ObjIds, obj_ids_to_list
from arcor2_arserver.lock.exceptions import CannotLock, LockingException
from arcor2_arserver.lock.structures import LockedObject, LockEventData
from arcor2_arserver.object_types.data import ObjectTypeDict
from arcor2_arserver_data.rpc.lock import UpdateType


class Lock:
    class Owners(cmn.StrEnum):
        """Values that can be used as lock owner substitution."""

        SERVER: str = "SERVER"

    class SpecialValues(cmn.StrEnum):
        """Values for special locking cases e.g. global variable access."""

        SCENE: str = "SCENE"
        PROJECT: str = "PROJECT"
        RUNNING_ACTION: str = "ACTION"
        ADDING_OBJECT: str = "ADDING_OBJECT"

    class ErrMessages(cmn.StrEnum):
        """Lock general error messages."""

        SOMETHING_LOCKED: str = "There are locked objects"
        SOMETHING_LOCKED_IN_TREE: str = "Part of tree is locked"
        LOCK_FAIL: str = "Locking failed, try again"
        NOT_LOCKED: str = "Object not locked"

    __slots__ = (
        "_scene",
        "_project",
        "_lock",
        "_locked_objects",
        "_release_tasks",
        "notifications_q",
        "_ui_user_locks",
        "_lock_timeout",
        "_lock_retries",
        "_retry_wait",
        "_object_types",
    )

    def __init__(self, obj_types: ObjectTypeDict) -> None:
        self._scene: None | UpdateableCachedScene = None
        self._project: None | UpdateableCachedProject = None
        self._object_types = obj_types

        self._lock: asyncio.Lock = asyncio.Lock()
        self._locked_objects: dict[str, LockedObject] = {}
        self._release_tasks: dict[str, asyncio.Task] = {}

        self.notifications_q: asyncio.Queue[LockEventData] = asyncio.Queue()
        self._ui_user_locks: dict[str, set[str]] = {}

        self._lock_timeout: float = env.get_float("ARCOR2_ARSERVER_LOCK_TIMEOUT", 300)
        self._lock_retries: int = env.get_int("ARCOR2_ARSERVER_LOCK_RETRIES", 13)
        self._retry_wait: float = env.get_float("ARCOR2_ARSERVER_LOCK_RETRY_WAIT", 0.15)

    @property
    def scene(self) -> None | UpdateableCachedScene:
        return self._scene

    @scene.setter
    def scene(self, scene: None | UpdateableCachedScene = None) -> None:
        self._scene = scene

    def scene_or_exception(self, ensure_project_closed: bool = False) -> UpdateableCachedScene:
        if not self._scene:
            raise Arcor2Exception("Scene not opened.")

        if ensure_project_closed and self._project:
            raise Arcor2Exception("Project has to be closed first.")

        return self._scene

    @property
    def project(self) -> None | UpdateableCachedProject:
        return self._project

    @project.setter
    def project(self, project: None | UpdateableCachedProject = None) -> None:
        self._project = project

    def project_or_exception(self, *, must_have_logic: bool = False) -> UpdateableCachedProject:
        if not self._project:
            raise Arcor2Exception("Project not opened.")

        if must_have_logic and not self._project.has_logic:
            raise Arcor2Exception("Project without logic.")

        return self._project

    @property
    def all_ui_locks(self) -> dict[str, set[str]]:
        """All lock grayed-out in editor in format user: {obj_ids}"""

        return self._ui_user_locks

    @asynccontextmanager
    async def get_lock(self, dry_run: bool = False) -> AsyncGenerator[None | asyncio.Lock, None]:
        """Get lock for data structure, method should be used for operation
        with whole scene/project e.g. saving project, where no changes should
        be made during this operation.

        :param dry_run: self._lock is not acquired/blocked when set
        """

        if self._ui_user_locks:
            # TODO turn warn into debug once the issue with "Locking failed" is gone
            if not dry_run:
                logger.warn(f"Locking failed... ui_user_locks: {self._ui_user_locks}")
            raise CannotLock(self.ErrMessages.LOCK_FAIL.value)

        i = 0
        yielded = False
        try:
            for _ in range(self._lock_retries):
                i += 1
                try:
                    async with self._lock:  # TODO maybe use some timeout here?
                        if self._get_write_locks_count():
                            raise CannotLock(self.ErrMessages.LOCK_FAIL.value)

                        if not dry_run:
                            yielded = True
                            yield self._lock
                    if dry_run:
                        yielded = True
                        yield None

                    break
                except CannotLock:
                    await asyncio.sleep(self._retry_wait)
        finally:
            if i > self._lock_retries * 0.25:
                logger.warn(f"Retry took {i * self._lock_timeout}")

            if not yielded:
                # TODO turn warn into debug once the issue with "Locking failed" is gone
                logger.warn(
                    f"Locking failed... locked_objects: {self._locked_objects}, ui_user_locks: {self._ui_user_locks}."
                )
                raise CannotLock(self.ErrMessages.LOCK_FAIL.value)

    async def read_lock(self, obj_ids: ObjIds, owner: str) -> bool:
        """Lock object ids by owner if all checks pass. No object is locked
        otherwise.

        :param obj_ids: object identifiers to be locked
        :param owner: lock owner name
        """

        obj_ids = obj_ids_to_list(obj_ids)
        roots = [await self.get_root_id(obj_id) for obj_id in obj_ids]

        async with self._lock:
            return self._read_lock(roots, obj_ids, owner)

    def _read_lock(self, roots: list[str], obj_ids: Iterable[str], owner: str) -> bool:
        """Private method when lock is already acquired."""

        assert self._lock.locked()

        locked: list[str] = []
        for i, obj_id in enumerate(obj_ids):

            if self._get_lock_record(roots[i]).read_lock(obj_id, owner):
                locked.append(obj_id)
            else:
                self._read_unlock(roots, locked, owner)
                return False
        return True

    async def read_unlock(self, obj_ids: ObjIds, owner: str) -> None:
        """Check real lock owner and delete records from lock database.

        :param obj_ids: object identifiers to be unlocked
        :param owner: lock owner name
        """

        obj_ids = obj_ids_to_list(obj_ids)

        roots = [await self.get_root_id(obj_id) for obj_id in obj_ids]

        async with self._lock:
            self._read_unlock(roots, obj_ids, owner)

    def _read_unlock(self, roots: list[str], obj_ids: list[str], owner: str) -> None:
        """Private method when lock is already acquired."""

        assert self._lock.locked()

        for i, obj_id in enumerate(obj_ids):
            try:
                self._get_lock_record(roots[i]).read_unlock(obj_id, owner)
            except LockingException:
                self._read_unlock(roots[i + 1 :], obj_ids[i + 1 :], owner)
                raise
            finally:
                self._check_and_clean_root(roots[i])

    async def write_lock(self, obj_ids: ObjIds, owner: str, lock_tree: bool = False, notify: bool = False) -> bool:
        """Lock object or list of objects for writing, possible to lock whole
        tree where object(s) is located.

        :param obj_ids: object or objects to be locked
        :param owner: name of lock owner
        :param lock_tree: boolean whether lock whole tree
        :param notify: if set, broadcast information about locked objects and store those objects in user lock database
        :return: boolean whether lock was successful
        """

        assert isinstance(obj_ids, str) and lock_tree or not lock_tree

        obj_ids = obj_ids_to_list(obj_ids)
        roots = [await self.get_root_id(obj_id) for obj_id in obj_ids]

        async with self._lock:
            ret = self._write_lock(roots, obj_ids, owner, lock_tree)

        if ret and notify:
            ui_locked = set(obj_ids)
            if lock_tree:
                for obj_id in roots:  # TODO lock only tree from affected object
                    ui_locked.update(self.get_all_children(obj_id))
                ui_locked.update(roots)
            self._upsert_user_locked_objects(owner, list(ui_locked))
            self.notifications_q.put_nowait(LockEventData(ui_locked, owner, True))
        return ret

    def _write_lock(self, roots: list[str], obj_ids: Iterable[str], owner: str, lock_tree: bool = False) -> bool:
        """Private method when lock is already acquired."""

        assert self._lock.locked()

        locked: list[str] = []
        for i, obj_id in enumerate(obj_ids):

            if self._get_lock_record(roots[i]).write_lock(obj_id, owner, lock_tree):
                locked.append(obj_id)
            else:
                self._write_unlock(roots, locked, owner)
                return False
        return True

    async def write_unlock(self, obj_ids: ObjIds, owner: str, notify: bool = False) -> None:
        """Check object lock real owner and remove it from lock database.

        :param obj_ids: object identifiers to be locked
        :param owner: lock owner name
        :param notify: if set, send-out notification about unlocked objects
        """

        obj_ids = obj_ids_to_list(obj_ids)

        roots = [await self.get_root_id(obj_id) for obj_id in obj_ids]

        async with self._lock:
            locked_trees = self._write_unlock(roots, obj_ids, owner)

        ui_locked = set(obj_ids)
        for obj_id, locked_tree in zip(roots, locked_trees):  # TODO lock only tree from affected object
            if locked_tree:
                ui_locked.update(self.get_all_children(obj_id))
                ui_locked.update(roots)

        # Always try to remove locked objects from database, so it won't get messy when deleting too many objects
        self._remove_user_locked_objects(owner, list(ui_locked))

        if notify:
            self.notifications_q.put_nowait(LockEventData(ui_locked, owner))

    def _write_unlock(self, roots: list[str], obj_ids: list[str], owner: str) -> list[bool]:
        """Private method when lock is already acquired."""

        assert self._lock.locked()

        ret: list[bool] = []
        for i, obj_id in enumerate(obj_ids):
            lock_record = self._get_lock_record(roots[i])
            ret.append(lock_record.tree)
            try:
                lock_record.write_unlock(obj_id, owner)

            except LockingException:
                self._write_unlock(roots[i + 1 :], obj_ids[i + 1 :], owner)
                raise
            finally:
                self._check_and_clean_root(roots[i])
        return ret

    async def is_write_locked(self, obj_id: str, owner: str, check_tree_locked: bool = False) -> bool:
        """Checks if object is write locked.

        :param obj_id: object identifiers to be locked
        :param owner: lock owner name
        :param check_tree_locked: if set, checks also if whole tree is locked
        """

        root_id = await self.get_root_id(obj_id)

        async with self._lock:
            return self._is_write_locked(root_id, obj_id, owner, check_tree_locked)

    def _is_write_locked(self, root_id: str, obj_id: str, owner: str, check_tree_locked: bool) -> bool:
        """Private method when lock is already acquired."""

        assert self._lock.locked()

        if lock_record := self._locked_objects.get(root_id):

            if obj_id in lock_record.write and owner == lock_record.write[obj_id]:
                if check_tree_locked:
                    return lock_record.tree
                return True

            if lock_record.tree:
                write_locks = lock_record.write.values()

                # When tree is locked, always 1 write lock occurs
                assert len(write_locks) == 1

                return owner == next(iter(write_locks))

        return False

    async def is_read_locked(self, obj_id: str, owner: str) -> bool:
        """Checks if object is locked for read.

        :param obj_id: object identifier to check
        :param owner: user name which must own lock record
        """

        root_id = await self.get_root_id(obj_id)

        async with self._lock:
            return self._is_read_locked(root_id, obj_id, owner)

    def _is_read_locked(self, root_id: str, obj_id: str, owner: str) -> bool:
        """Private method when lock is already acquired."""

        assert self._lock.locked()

        if lock_record := self._locked_objects.get(root_id):
            return obj_id in lock_record.read and owner in lock_record.read[obj_id]
        return False

    async def get_locked_roots_count(self) -> int:
        """Count and return number of total locked roots."""

        async with self._lock:
            return len(self._locked_objects)

    async def get_write_locks_count(self) -> int:
        """Count and return total number of write locks."""

        async with self._lock:
            return self._get_write_locks_count()

    def _get_write_locks_count(self) -> int:
        """Private method when lock is already acquired."""

        assert self._lock.locked()

        return sum(len(lock.write) for lock in self._locked_objects.values())

    async def get_root_id(self, obj_id: str) -> str:
        """Retrieve root object id for given object. Works also for project and
        scene ID.

        :param obj_id: object to search root for
        """

        if obj_id in self.SpecialValues.set() | self._object_types.keys():
            return obj_id
        elif self.project and self.scene:
            if (
                obj_id
                in {self.scene.id, self.project.id, cmn.LogicItem.START, cmn.LogicItem.END}
                | self.project.parameters_ids
            ):
                return obj_id
            elif obj_id in self.scene.object_ids:
                # TODO implement with scene object hierarchy
                return obj_id
            else:
                parent = self.project.get_parent_id(obj_id)
                ret = parent if parent else obj_id
                try:
                    while parent:
                        parent = self.project.get_parent_id(parent)
                        if parent and parent != ret:
                            ret = parent
                except CachedProjectException:
                    ...
                return ret

        elif self.scene and obj_id in self.scene.object_ids | {self.scene.id}:
            # TODO implement with scene object hierarchy
            return obj_id

        # locking on dashboard, check if scene or project exists
        if obj_id in set.union(*await asyncio.gather(storage.get_scene_ids(), storage.get_project_ids())):
            return obj_id

        raise Arcor2Exception(f"Unknown object '{obj_id}'.")

    def _get_lock_record(self, root_id: str) -> LockedObject:
        """Create and/or retrieve lock record for root_id."""

        assert self._lock.locked()

        if root_id not in self._locked_objects:
            self._locked_objects[root_id] = LockedObject()

        return self._locked_objects[root_id]

    async def update_write_lock(self, locked_obj_id: str, parent_id: str, owner: str):
        """Update lock record of locked object to new root, both objects
        expected to be locked.

        :param locked_obj_id: ID of locked object to be updated
        :param parent_id: ID of previous parent of locked object
        :param owner: name of lock owner
        """

        new_root_id = await self.get_root_id(locked_obj_id)

        async with self._lock:
            self._write_unlock([parent_id], [locked_obj_id], owner)
            self._write_lock([new_root_id], [locked_obj_id], owner)

    def _check_and_clean_root(self, root: str) -> None:
        """Removes top level lock record if read and write attributes are
        empty."""

        assert self._lock.locked()

        if self._locked_objects[root].is_empty():
            del self._locked_objects[root]

    async def schedule_auto_release(self, owner: str) -> None:
        """Creates task for releasing all user locks.

        :param owner: affected user
        """

        self._release_tasks[owner] = asyncio.create_task(self._release_all_owner_locks(owner))

    async def cancel_auto_release(self, owner: str) -> None:
        """Cancel task releasing user locks after timeout. Used after login.

        :param owner: affected user name
        """

        if owner in self._release_tasks:
            self._release_tasks[owner].cancel()
            del self._release_tasks[owner]

    async def _release_all_owner_locks(self, owner: str) -> None:
        """Task planned after user logout. Release all its lock after timeout
        and notify UIs.

        :param owner: unregistered user
        """

        await asyncio.sleep(self._lock_timeout)

        async with self._lock:
            read, write = self._get_owner_locks(owner)

            if read or write:
                logger.warn(f"{len(read)+len(write)} lock(s) of {owner} were just discarded.")

            self._read_unlock(list(read), list(read.values()), owner)
            self._write_unlock(list(write), list(write.values()), owner)

        if to_notify := self._ui_user_locks.pop(owner):
            self.notifications_q.put_nowait(LockEventData(to_notify, owner))

    async def get_owner_locks(self, owner: str) -> tuple[dict[str, str], dict[str, str]]:
        """Finds which locks belongs to owner.

        :param owner: user name
        :return: tuple of read and write locks in format object_id: root id
        """

        async with self._lock:
            return self._get_owner_locks(owner)

    def _get_owner_locks(self, owner: str) -> tuple[dict[str, str], dict[str, str]]:
        """Private method when lock is already acquired."""

        assert self._lock.locked()

        read: dict[str, str] = {}
        write: dict[str, str] = {}

        for root, data in self._locked_objects.items():
            for obj_id, owners in data.read.items():
                if owner in owners:
                    read[obj_id] = root
            for obj_id, w_owner in data.write.items():
                if owner == w_owner:
                    write[obj_id] = root

        return read, write

    def _upsert_user_locked_objects(self, owner: str, obj_ids: Iterable[str]) -> None:
        """Add objects where lock is visible in UI to related database.

        :param owner: user name of locks owner
        :param obj_ids: objects to be added to user lock database
        """

        if owner not in self._ui_user_locks:
            self._ui_user_locks[owner] = set()

        self._ui_user_locks[owner].update(obj_ids)

    def _remove_user_locked_objects(self, owner: str, obj_ids: Iterable[str]) -> None:
        """Remove items where lock is visible in UI from related database.

        :param owner: user name of locks owner
        :param obj_ids: objects to be removed from user lock database
        """

        if owner in self._ui_user_locks:
            self._ui_user_locks[owner].difference_update(obj_ids)

            if not self._ui_user_locks[owner]:
                del self._ui_user_locks[owner]

    def get_all_children(self, obj_id: str) -> set[str]:
        """Recursively find all children of object.

        :param obj_id: parent of all found children
        """

        return self.project.childs(obj_id, recursive=True) if self.project else set()

    def get_all_parents(self, obj_id: str) -> set[str]:
        """Recursively find all parents in tree.

        :param obj_id: object to search parents for
        """

        parents: set[str] = set()
        if self.project:
            parent = self.project.get_parent_id(obj_id)
            while parent:
                parents.add(parent)
                parent = self.project.get_parent_id(parent)
        elif self.scene:
            ...  # TODO implement with scene object hierarchy

        return parents

    async def check_remove(self, obj_id: str, owner: str) -> bool:
        """Check if object can be removed, e.g. no child locked, not in locked
        tree.

        :param obj_id: object id to be checked
        :param owner: user who tries to remove object, returns true if user owns affected lock
        """

        def check_children(_root: str, obj_ids: set[str], read: bool = False) -> bool:
            lock_item = self._locked_objects[_root]
            for _obj_id in obj_ids:
                if read:
                    if owner not in lock_item.read[_obj_id]:
                        return False
                    if lock_item.read[_obj_id].count != 1:
                        return False
                else:
                    if owner != lock_item.write[_obj_id]:
                        return False
            return True

        root_id = await self.get_root_id(obj_id)
        children = self.get_all_children(obj_id)
        children.add(obj_id)

        async with self._lock:
            if lock_record := self._locked_objects.get(root_id):

                if lock_record.tree and owner != next(iter(lock_record.write.values())):
                    return False

                if not check_children(root_id, children.intersection(lock_record.write)):
                    return False
                if not check_children(root_id, children.intersection(lock_record.read), True):
                    return False
                return True

            return False

    TypesWithId = (
        cmn.SceneObject
        | cmn.BareActionPoint
        | cmn.NamedOrientation
        | cmn.ProjectRobotJoints
        | cmn.Action
        | cmn.ProjectParameter
    )

    def get_by_id(self, obj_id: str) -> TypesWithId:
        """Retrive object by it's ID."""

        if self.project:
            try:
                return self.project.get_by_id(obj_id)
            except CachedProjectException:
                ...

        if self.scene:  # TODO update with scene object hierarchy
            if obj_id in self.scene.object_ids:
                return self.scene.object(obj_id)

        raise Arcor2Exception(f"Object ID {obj_id} not found.")

    async def update_lock(self, obj_id: str, owner: str, upgrade_type: UpdateType) -> None:
        """Upgrades lock to locked whole tree or downgrades lock to simple
        object lock.

        :param obj_id: objects which is locked and updated
        :param owner: owner of current lock
        :param upgrade_type: one of available type
        """

        root_id = await self.get_root_id(obj_id)

        async with self._lock:
            if root_id not in self._locked_objects:
                raise LockingException(self.ErrMessages.NOT_LOCKED.value)

            lock_record = self._get_lock_record(root_id)
            if upgrade_type == UpdateType.TREE:
                lock_record.check_upgrade(obj_id, owner)
                lock_record.tree = True

                to_notify = self.get_all_children(root_id)
                to_notify.add(root_id)
                to_notify.remove(obj_id)
                evt = LockEventData(to_notify, owner, True)

                self._upsert_user_locked_objects(owner, to_notify)
            elif upgrade_type == UpdateType.OBJECT:
                lock_record.check_downgrade(obj_id, owner)
                lock_record.tree = False

                to_notify = self.get_all_children(root_id)
                to_notify.add(root_id)
                to_notify -= {obj_id}
                evt = LockEventData(to_notify, owner)

                self._remove_user_locked_objects(owner, to_notify)
            else:
                raise Arcor2Exception("Unknown type of lock upgrade")

            self.notifications_q.put_nowait(evt)

    async def check_lock_tree(self, obj_id: str, owner: str) -> None:
        """Checks if locking whole tree is possible.

        :param obj_id: object id to dry run locking for
        :param owner: owner of dry run lock
        """

        root_id = await self.get_root_id(obj_id)
        children = self.get_all_children(root_id)
        children.add(root_id)

        async with self._lock:
            if lock_record := self._locked_objects.get(root_id):

                for lock in children.intersection(lock_record.write):
                    if owner != lock_record.write[lock]:
                        raise CannotLock(self.ErrMessages.SOMETHING_LOCKED_IN_TREE.value)

                for lock in children.intersection(lock_record.read):
                    if len(lock_record.read[lock]) > 1 or owner not in lock_record.read[lock]:
                        raise CannotLock(self.ErrMessages.SOMETHING_LOCKED_IN_TREE.value)



import asyncio
import time
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any, Awaitable, Callable

from lru import LRU

from arcor2 import env
from arcor2.cached import CachedProject, CachedScene
from arcor2.clients import aio_project_service as ps
from arcor2.clients.aio_project_service import (
    delete_model,
    get_mesh,
    get_meshes,
    get_model,
    get_project_sources,
    put_model,
    update_project_sources,
)
from arcor2.clients.project_service import ProjectServiceException
from arcor2.data.common import IdDesc
from arcor2.data.object_type import ObjectType
from arcor2.exceptions import Arcor2Exception


@dataclass
class CachedListing:

    __slots__ = "listing", "ts"

    listing: dict[str, IdDesc]
    ts: float

    def time_to_update(self) -> bool:
        return time.monotonic() - self.ts > _cache_timeout


"""
This module adds some caching capabilities to the aio version of project_service. It should be only used by ARServer.

Caching can be disabled by setting respective environment variable - this is useful for environments
where ARServer is not the only one who touches Project service.
"""

_cache_timeout = max(env.get_float("ARCOR2_ARSERVER_CACHE_TIMEOUT", 1.0), 0)
_cache_scenes = max(env.get_int("ARCOR2_ARSERVER_CACHE_SCENES", 32), 1)
_cache_projects = max(env.get_int("ARCOR2_ARSERVER_CACHE_PROJECTS", 64), 1)
_cache_object_types = max(env.get_int("ARCOR2_ARSERVER_CACHE_OBJECT_TYPES", 64), 1)

# here we need to know all the items
_scenes_list = CachedListing({}, 0)
_projects_list = CachedListing({}, 0)
_object_type_list = CachedListing({}, 0)

_scenes_list_lock = asyncio.Lock()
_projects_list_lock = asyncio.Lock()
_object_type_lock = asyncio.Lock()

# here we can forget the least used items
if TYPE_CHECKING:
    _scenes: dict[str, CachedScene] = {}
    _projects: dict[str, CachedProject] = {}
    _object_types: dict[str, ObjectType] = {}
else:
    _scenes = LRU(_cache_scenes)
    _projects = LRU(_cache_projects)
    _object_types = LRU(_cache_object_types)


async def _update_list(
    getter: Callable[..., Awaitable[list[IdDesc]]], cached_listing: CachedListing, cache: dict[str, Any]
) -> None:

    if not cached_listing.time_to_update():
        return

    updated = {it.id: it for it in (await getter())}
    for deleted in cached_listing.listing.keys() - updated.keys():  # remove outdated items from the cache
        cache.pop(deleted, None)
    cached_listing.listing = updated
    cached_listing.ts = time.monotonic()


async def initialize_module() -> None:

    await asyncio.gather(
        _update_list(ps.get_projects, _projects_list, _projects),
        _update_list(ps.get_scenes, _scenes_list, _scenes),
        _update_list(ps.get_object_type_ids, _object_type_list, _object_types),
    )

    _scenes.clear()
    _projects.clear()
    _object_types.clear()


async def get_project_ids() -> set[str]:

    async with _projects_list_lock:
        await _update_list(ps.get_projects, _projects_list, _projects)
    return set(_projects_list.listing)


async def get_projects() -> list[IdDesc]:

    async with _projects_list_lock:
        await _update_list(ps.get_projects, _projects_list, _projects)
    return list(_projects_list.listing.values())


async def get_scene_ids() -> set[str]:

    async with _scenes_list_lock:
        await _update_list(ps.get_scenes, _scenes_list, _scenes)
    return set(_scenes_list.listing)


async def get_scenes() -> list[IdDesc]:

    async with _scenes_list_lock:
        await _update_list(ps.get_scenes, _scenes_list, _scenes)
    return list(_scenes_list.listing.values())


async def get_object_type_ids() -> set[str]:

    async with _object_type_lock:
        await _update_list(ps.get_object_type_ids, _object_type_list, _object_types)
    return set(_object_type_list.listing)


async def get_object_types() -> list[IdDesc]:

    async with _object_type_lock:
        await _update_list(ps.get_object_type_ids, _object_type_list, _object_types)
    return list(_object_type_list.listing.values())


async def get_object_type_iddesc(object_type_id: str) -> IdDesc:

    async with _object_type_lock:
        await _update_list(ps.get_object_type_ids, _object_type_list, _object_types)
    try:
        return _object_type_list.listing[object_type_id]
    except KeyError:
        raise ProjectServiceException("Unknown object type.")


async def get_project(project_id: str) -> CachedProject:

    async with _projects_list_lock:
        try:
            project = _projects[project_id]
            assert project.modified
        except KeyError:
            project = CachedProject(await ps.get_project(project_id))
            _projects[project_id] = project
        else:
            await _update_list(ps.get_projects, _projects_list, _projects)

            if project_id not in _projects_list.listing:
                _projects.pop(project_id, None)
                raise Arcor2Exception("Project removed externally.")

            # project in cache is outdated
            if project.modified < _projects_list.listing[project_id].modified:
                project = CachedProject(await ps.get_project(project_id))
                _projects[project_id] = project

    return project


async def get_scene(scene_id: str) -> CachedScene:

    async with _scenes_list_lock:
        try:
            scene = _scenes[scene_id]
            assert scene.modified
        except KeyError:
            scene = CachedScene(await ps.get_scene(scene_id))
            _scenes[scene_id] = scene
        else:
            await _update_list(ps.get_scenes, _scenes_list, _scenes)

            if scene_id not in _scenes_list.listing:
                _scenes.pop(scene_id, None)
                raise Arcor2Exception("Scene removed externally.")

            # scene in cache is outdated
            if scene.modified < _scenes_list.listing[scene_id].modified:
                scene = CachedScene(await ps.get_scene(scene_id))
                _scenes[scene_id] = scene

    return scene


async def get_object_type(object_type_id: str) -> ObjectType:

    async with _object_type_lock:
        try:
            ot = _object_types[object_type_id]
            assert ot.modified
        except KeyError:
            ot = await ps.get_object_type(object_type_id)
            _object_types[object_type_id] = ot
        else:
            await _update_list(ps.get_object_type_ids, _object_type_list, _object_types)

            if object_type_id not in _object_type_list.listing:
                _object_types.pop(object_type_id, None)
                raise Arcor2Exception("ObjectType removed externally.")

            # ObjectType in cache is outdated
            if ot.modified < _object_type_list.listing[object_type_id].modified:
                ot = await ps.get_object_type(object_type_id)
                _object_types[object_type_id] = ot

    return ot


async def update_project(project: CachedProject) -> datetime:

    assert project.id

    ret = await ps.update_project(project.project)
    project.modified = ret

    if not project.created:
        project.created = project.modified

    _projects_list.listing[project.id] = IdDesc(
        project.id, project.name, project.created, project.modified, project.description
    )
    _projects[project.id] = deepcopy(project)
    _projects[project.id].int_modified = None

    return ret


async def update_scene(scene: CachedScene) -> datetime:

    assert scene.id

    ret = await ps.update_scene(scene.scene)
    scene.modified = ret

    if not scene.created:
        scene.created = scene.modified

    _scenes_list.listing[scene.id] = IdDesc(scene.id, scene.name, scene.created, scene.modified, scene.description)
    _scenes[scene.id] = deepcopy(scene)
    _scenes[scene.id].int_modified = None

    return ret


async def update_object_type(object_type: ObjectType) -> datetime:

    assert object_type.id

    ret = await ps.update_object_type(object_type)
    object_type.modified = ret

    if not object_type.created:
        object_type.created = object_type.modified

    _object_type_list.listing[object_type.id] = IdDesc(
        object_type.id, "", object_type.created, object_type.modified, object_type.description
    )
    _object_types[object_type.id] = deepcopy(object_type)

    return ret


async def delete_scene(scene_id: str) -> None:

    await ps.delete_scene(scene_id)
    _scenes_list.listing.pop(scene_id, None)
    _scenes.pop(scene_id, None)


async def delete_project(project_id: str) -> None:

    await ps.delete_project(project_id)
    _projects_list.listing.pop(project_id, None)
    _projects.pop(project_id, None)


async def delete_object_type(object_type_id: str) -> None:

    await ps.delete_object_type(object_type_id)
    _object_type_list.listing.pop(object_type_id, None)
    _object_types.pop(object_type_id, None)


__all__ = [
    initialize_module.__name__,
    get_mesh.__name__,
    get_meshes.__name__,
    get_model.__name__,
    put_model.__name__,
    delete_model.__name__,
    get_projects.__name__,
    get_scenes.__name__,
    get_project.__name__,
    get_project_sources.__name__,
    get_scene.__name__,
    get_object_type.__name__,
    get_object_types.__name__,
    update_project.__name__,
    update_scene.__name__,
    update_project_sources.__name__,
    update_object_type.__name__,
    delete_object_type.__name__,
    delete_scene.__name__,
    delete_project.__name__,
    ProjectServiceException.__name__,
    get_scene_ids.__name__,
    get_project_ids.__name__,
    get_object_type_ids.__name__,
]



import asyncio
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable

from arcor2 import helpers as hlp
from arcor2.cached import CachedProject, CachedScene, UpdateableCachedProject
from arcor2.exceptions import Arcor2Exception
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver.checks import project_problems
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.objects_actions import get_object_types
from arcor2_arserver.scene import SceneProblems, get_ot_modified, get_scene_state, open_scene
from arcor2_arserver_data.events.actions import ActionExecution, ActionResult
from arcor2_arserver_data.events.common import ShowMainScreen
from arcor2_arserver_data.events.project import OpenProject, ProjectClosed
from arcor2_runtime.action import results_to_json


@dataclass
class ProjectProblems(SceneProblems):
    project_modified: datetime


_project_problems: dict[str, ProjectProblems] = {}


async def get_project_problems(scene: CachedScene, project: CachedProject) -> None | list[str]:
    """Handle caching of project problems."""

    assert scene.modified
    assert project.modified

    ots = scene.object_types
    await get_object_types()
    ot_modified = get_ot_modified(ots)

    if (
        project.id not in _project_problems
        or _project_problems[project.id].scene_modified < scene.modified
        or _project_problems[project.id].project_modified < project.modified
        or _project_problems[project.id].ot_modified != ot_modified
    ):

        logger.debug(f"Updating project_problems for {project.name}.")

        _project_problems[project.id] = ProjectProblems(
            scene.modified,
            project_problems(glob.OBJECT_TYPES, scene, project),
            ot_modified,
            project.modified,
        )

    # prune removed projects
    for csi in set(_project_problems.keys()) - await storage.get_project_ids():
        logger.debug(f"Pruning cached problems for removed project {csi}.")
        _project_problems.pop(csi, None)

    sp = _project_problems[project.id].problems

    return sp if sp else None


async def notify_project_opened(evt: OpenProject) -> None:

    await notif.broadcast_event(evt)
    await notif.broadcast_event(get_scene_state())


async def notify_project_closed(project_id: str, show_mainscreen_after_that: bool = True) -> None:

    proj_list = ShowMainScreen.Data.WhatEnum.ProjectsList

    await notif.broadcast_event(ProjectClosed())
    if show_mainscreen_after_that:  # mainscreen is not shown when running a temporary package
        glob.MAIN_SCREEN = ShowMainScreen.Data(proj_list)
        await notif.broadcast_event(ShowMainScreen(ShowMainScreen.Data(proj_list, project_id)))


async def close_project(show_mainscreen_after_that: bool = True) -> None:

    assert glob.LOCK.project

    project_id = glob.LOCK.project.project.id
    glob.LOCK.scene = None
    glob.LOCK.project = None
    glob.PREV_RESULTS.clear()
    asyncio.ensure_future(notify_project_closed(project_id, show_mainscreen_after_that))


async def execute_action(action_method: Callable, params: list[Any]) -> None:

    assert glob.RUNNING_ACTION

    await notif.broadcast_event(ActionExecution(ActionExecution.Data(glob.RUNNING_ACTION)))

    evt = ActionResult(ActionResult.Data(glob.RUNNING_ACTION))

    try:
        action_result = await hlp.run_in_executor(action_method, *params)
    except Arcor2Exception as e:
        logger.error(f"Failed to run method {action_method.__name__} with params {params}. {str(e)}")
        logger.debug(str(e), exc_info=True)
        evt.data.error = str(e)
    else:

        if action_result is not None:

            glob.PREV_RESULTS[glob.RUNNING_ACTION] = action_result

            try:
                evt.data.results = results_to_json(action_result)
            except Arcor2Exception:
                logger.error(f"Method {action_method.__name__} returned unsupported type of result: {action_result}.")

    if glob.RUNNING_ACTION is None:
        # action was cancelled, do not send any event
        return  # type: ignore  # action could be cancelled during its execution

    glob.RUNNING_ACTION = None
    glob.RUNNING_ACTION_PARAMS = None
    await notif.broadcast_event(evt)


async def open_project(project_id: str) -> None:

    project = await storage.get_project(project_id)

    if glob.LOCK.scene:
        if glob.LOCK.scene.id != project.scene_id:
            raise Arcor2Exception("Required project is associated to another scene.")
    else:
        await open_scene(project.scene_id)

    assert glob.LOCK.scene

    if pp := await get_project_problems(glob.LOCK.scene, project):
        glob.LOCK.scene = None
        logger.warning(f"Project {project.name} can't be opened due to the following problem(s)...")
        for ppp in pp:
            logger.warning(ppp)
        raise Arcor2Exception("Project has some problems.")

    glob.LOCK.project = UpdateableCachedProject(project)



import asyncio

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver.lock.structures import LockEventData
from arcor2_arserver.rpc.objects import object_aiming_prune
from arcor2_arserver_data import rpc as srpc


async def register_user_cb(req: srpc.u.RegisterUser.Request, ui: WsClient) -> None:

    await glob.USERS.login(req.args.user_name, ui)
    logger.debug(f"User {req.args.user_name} just logged in. Known user names are: {glob.USERS.user_names}")

    await glob.LOCK.cancel_auto_release(req.args.user_name)

    # those are locks that are known for all users
    for user, user_objects in glob.LOCK.all_ui_locks.items():
        if user != req.args.user_name:
            glob.LOCK.notifications_q.put_nowait(LockEventData(user_objects, user, True, ui))

    # these locks are known only to the current user
    if user_write_locks := (await glob.LOCK.get_owner_locks(req.args.user_name))[1]:
        glob.LOCK.notifications_q.put_nowait(LockEventData(user_write_locks, req.args.user_name, True, ui))

    asyncio.create_task(object_aiming_prune())

    return None



from typing import Iterable

ObjIds = Iterable[str] | str


def obj_ids_to_list(val: ObjIds) -> list[str]:

    if isinstance(val, str):
        return [val]
    elif isinstance(val, list):
        return val
    else:
        return list(val)



import asyncio

from websockets.exceptions import WebSocketException
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.exceptions import Arcor2Exception


class UsersException(Arcor2Exception):
    pass


class Users:

    __slots__ = ("_interfaces", "_users_ui", "_ui_users")

    def __init__(self) -> None:

        self._interfaces: set[WsClient] = set()
        self._users_ui: dict[str, WsClient] = {}
        self._ui_users: dict[WsClient, str] = {}

    def _assert_consistency(self) -> None:

        assert list(self._ui_users.keys()) == list(self._users_ui.values())
        assert list(self._ui_users.values()) == list(self._users_ui.keys())
        assert set(self._ui_users).issubset(self._interfaces)

    @property
    def interfaces(self) -> set[WsClient]:
        return self._interfaces

    def add_interface(self, ui: WsClient) -> None:
        self._interfaces.add(ui)

    async def login(self, user_name: str, ui: WsClient) -> None:

        self._assert_consistency()

        if not user_name:
            raise UsersException("Empty user name.")

        if ui not in self._interfaces:
            raise UsersException("Unknown ui.")

        if user_name in self._users_ui:

            old_ui = self._users_ui[user_name]

            try:
                pong_waiter = await old_ui.ping()
                await asyncio.wait_for(pong_waiter, 1.0)
            except (WebSocketException, asyncio.TimeoutError):
                self._users_ui.pop(user_name, None)
                self._ui_users.pop(old_ui, None)
            else:
                raise UsersException("Username already exists")

        self._ui_users[ui] = user_name
        self._users_ui[user_name] = ui

        self._assert_consistency()

    def logout(self, ui: WsClient) -> None:
        """Removes ui from known interfaces and logs a user out (if logged in).

        :param ui:
        :return:
        """

        self._assert_consistency()

        try:
            self._interfaces.remove(ui)
        except KeyError:
            raise UsersException("Unknown ui.")

        if user_name := self._ui_users.pop(ui, None):
            self._users_ui.pop(user_name, None)

        self._assert_consistency()

    @property
    def user_names(self) -> set[str]:
        return set(self._users_ui)

    def user_name(self, ui: WsClient) -> str:

        self._assert_consistency()

        try:
            return self._ui_users[ui]
        except KeyError:
            raise UsersException("User not logged in.")



import functools
from asyncio import sleep
from typing import Any, Callable, Coroutine, TypeVar, cast

from arcor2_arserver import logger
from arcor2_arserver.lock.exceptions import LockingException

F = TypeVar("F", bound=Callable[..., Coroutine[Any, Any, Any]])


def retry(exc: type[Exception] = LockingException, tries: int = 1, delay: float = 0) -> Callable:
    """Retry decorator.

    :param exc: Exception or tuple of exceptions to catch
    :param tries: number of attempts
    :param delay: delay between attempts
    :return:
    """

    async def _retry_call(_tries: int, coro: F, *args, **kwargs) -> Any:
        """
        Internal part of retry decorator - handles exceptions, delay and tries
        """
        while _tries:
            try:
                ret = await coro(*args, **kwargs)
                if _tries < tries * 0.75:
                    logger.warn(f"Retry timeout took {(tries - _tries) * delay}")
                return ret
            except exc:
                _tries -= 1
                if _tries == 0:
                    logger.warn("Retry raised")
                    raise

                if delay > 0:
                    await sleep(delay)

    def _retry(coro: F) -> F:
        @functools.wraps(coro)
        async def async_wrapper(*fargs, **fkwargs) -> Any:
            return await _retry_call(tries, coro, *fargs, **fkwargs)

        return cast(F, async_wrapper)

    return _retry



import asyncio
import copy
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2 import transformations as tr
from arcor2.cached import CachedProject, UpdateableCachedProject
from arcor2.data import common
from arcor2.data.events import Event, PackageState
from arcor2.exceptions import Arcor2Exception
from arcor2.logic import check_for_loops
from arcor2.object_types.abstract import Generic, Robot
from arcor2.parameter_plugins.base import ParameterPluginException
from arcor2.parameter_plugins.pose import PosePlugin
from arcor2.parameter_plugins.utils import plugin_from_type_name
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver import project
from arcor2_arserver.checks import (
    check_action_params,
    check_ap_parent,
    check_flows,
    check_logic_item,
    check_project_parameter,
    find_object_action,
)
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.common import project_names, scene_names
from arcor2_arserver.helpers import (
    ctx_read_lock,
    ctx_write_lock,
    ensure_write_locked,
    get_unlocked_objects,
    make_name_unique,
    unique_name,
)
from arcor2_arserver.objects_actions import get_types_dict
from arcor2_arserver.project import close_project, get_project_problems, notify_project_opened, open_project
from arcor2_arserver.robot import check_eef_arm, get_end_effector_pose, get_pose_and_joints, get_robot_joints
from arcor2_arserver.scene import (
    can_modify_scene,
    ensure_scene_started,
    get_instance,
    get_robot_instance,
    open_scene,
    save_scene,
)
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data import rpc as srpc


@asynccontextmanager
async def managed_project(project_id: str, make_copy: bool = False) -> AsyncGenerator[UpdateableCachedProject, None]:

    save_back = False

    if glob.LOCK.project and glob.LOCK.project.id == project_id:
        if make_copy:
            project = copy.deepcopy(glob.LOCK.project)
            save_back = True
        else:
            project = glob.LOCK.project
    else:
        save_back = True
        project = UpdateableCachedProject(await storage.get_project(project_id))

    if make_copy:
        project.id = common.Project.uid()

    try:
        yield project
    finally:
        if save_back:
            asyncio.ensure_future(storage.update_project(project))


async def cancel_action_cb(req: srpc.p.CancelAction.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    if not glob.RUNNING_ACTION:
        raise Arcor2Exception("No action is running.")

    action = proj.action(glob.RUNNING_ACTION)
    obj_id, action_name = action.parse_type()
    obj = get_instance(obj_id, Generic)

    if not find_object_action(glob.OBJECT_TYPES, scene, action).meta.cancellable:
        raise Arcor2Exception("Action is not cancellable.")

    try:
        action_method = getattr(obj, action_name)
        cancel_method = getattr(obj, obj.CANCEL_MAPPING[action_method.__name__])
    except AttributeError as e:
        raise Arcor2Exception("Internal error.") from e

    cancel_params: dict[str, Any] = {}

    # TODO fix or remove this?
    """
    cancel_sig = inspect.signature(cancel_method)

    assert glob.RUNNING_ACTION_PARAMS is not None

    for param_name in cancel_sig.parameters.keys():
        try:
            cancel_params[param_name] = glob.RUNNING_ACTION_PARAMS[param_name]
        except KeyError as e:
            raise Arcor2Exception("Cancel method parameters should be subset of action parameters.") from e
    """

    await hlp.run_in_executor(cancel_method, *cancel_params.values())

    asyncio.ensure_future(notif.broadcast_event(sevts.a.ActionCancelled()))
    glob.RUNNING_ACTION = None
    glob.RUNNING_ACTION_PARAMS = None


async def execute_action_cb(req: srpc.p.ExecuteAction.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    # TODO rather lock the project and release the lock once execution is finished?
    async with ctx_write_lock(glob.LOCK.SpecialValues.RUNNING_ACTION, glob.USERS.user_name(ui)):

        ensure_scene_started()

        if glob.RUNNING_ACTION:
            raise Arcor2Exception(
                f"Action {glob.RUNNING_ACTION} is being executed. " f"Only one action can be executed at a time."
            )

        action = proj.action(req.args.action_id)

        obj_id, action_name = action.parse_type()

        params: list[Any] = []

        for param in action.parameters:

            if param.type == common.ActionParameter.TypeEnum.LINK:

                parsed_link = param.parse_link()
                try:
                    results = glob.PREV_RESULTS[parsed_link.action_id]
                except KeyError:
                    prev_action = proj.action(parsed_link.action_id)
                    raise Arcor2Exception(f"Action '{prev_action.name}' has to be executed first.")

                # an action result could be a tuple or a single value
                if isinstance(results, tuple):
                    params.append(results[parsed_link.output_index])
                else:
                    assert parsed_link.output_index == 0
                    params.append(results)

            elif param.type == common.ActionParameter.TypeEnum.PROJECT_PARAMETER:
                pparam = proj.parameter(param.str_from_value())
                # TODO use plugin to get the value
                from arcor2 import json

                params.append(json.loads(pparam.value))
            else:

                try:
                    params.append(
                        plugin_from_type_name(param.type).parameter_execution_value(
                            get_types_dict(), scene, proj, action.id, param.name
                        )
                    )
                except ParameterPluginException as e:
                    logger.error(e)
                    raise Arcor2Exception(f"Failed to get value for parameter {param.name}.")

        obj = get_instance(obj_id, Generic)

        if isinstance(obj, Robot):
            if obj.move_in_progress:
                raise Arcor2Exception("Can't execute actions while the robot moves.")

        if not hasattr(obj, action_name):
            raise Arcor2Exception("Internal error: object does not have the requested method.")

        if req.dry_run:
            return

        glob.RUNNING_ACTION = action.id
        glob.RUNNING_ACTION_PARAMS = params

        logger.debug(f"Running action {action.name} ({type(obj)}/{action_name}), params: {params}.")

        # schedule execution and return success
        asyncio.ensure_future(project.execute_action(getattr(obj, action_name), params))
        return None


async def list_projects_cb(req: srpc.p.ListProjects.Request, ui: WsClient) -> srpc.p.ListProjects.Response:
    async def project_info(project_id: str) -> srpc.p.ListProjects.Response.Data:

        project = await storage.get_project(project_id)

        assert project.created
        assert project.modified

        pd = srpc.p.ListProjects.Response.Data(
            project.name,
            project.scene_id,
            project.description,
            project.has_logic,
            project.created,
            project.modified,
            id=project.id,
        )

        try:
            scene = await storage.get_scene(project.scene_id)
        except storage.ProjectServiceException:
            pd.problems = ["Scene does not exist."]
            return pd

        pd.problems = await get_project_problems(scene, project)
        return pd

    resp = srpc.p.ListProjects.Response()

    resp.data = []
    for res in await asyncio.gather(
        *[project_info(proj_id) for proj_id in (await storage.get_project_ids())], return_exceptions=True
    ):

        if isinstance(res, Arcor2Exception):
            logger.error(str(res))
        elif isinstance(res, Exception):
            raise res  # zero toleration for other exceptions
        else:
            resp.data.append(res)

    return resp


async def get_project_cb(req: srpc.p.GetProject.Request, ui: WsClient) -> srpc.p.GetProject.Response:

    return srpc.p.GetProject.Response(data=(await storage.get_project(req.args.id)).project)


async def add_ap_using_robot_cb(req: srpc.p.AddApUsingRobot.Request, ui: WsClient) -> None:
    async def notify(ap: common.BareActionPoint, ori: common.NamedOrientation, joi: common.ProjectRobotJoints) -> None:

        ap_evt = sevts.p.ActionPointChanged(ap)
        ap_evt.change_type = Event.Type.ADD
        await notif.broadcast_event(ap_evt)

        ori_evt = sevts.p.OrientationChanged(ori)
        ori_evt.change_type = Event.Type.ADD
        ori_evt.parent_id = ap.id

        joi_evt = sevts.p.JointsChanged(joi)
        joi_evt.change_type = Event.Type.ADD
        joi_evt.parent_id = ap.id

        await asyncio.gather(notif.broadcast_event(ori_evt), notif.broadcast_event(joi_evt))

    hlp.is_valid_identifier(req.args.name)
    proj = glob.LOCK.project_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):

        ensure_scene_started()

        unique_name(req.args.name, proj.action_points_names)

        robot_inst = get_robot_instance(req.args.robot_id)
        await check_eef_arm(robot_inst, req.args.arm_id, req.args.end_effector_id)

        if req.dry_run:
            return None

        pose, joints = await get_pose_and_joints(robot_inst, req.args.end_effector_id, req.args.arm_id)

        ap = proj.upsert_action_point(common.ActionPoint.uid(), req.args.name, pose.position)
        ori = common.NamedOrientation("default", pose.orientation)
        proj.upsert_orientation(ap.id, ori)
        joi = common.ProjectRobotJoints(
            "default", req.args.robot_id, joints, True, req.args.arm_id, req.args.end_effector_id
        )
        proj.upsert_joints(ap.id, joi)

        asyncio.ensure_future(notify(ap, ori, joi))
        return None


async def add_action_point_joints_using_robot_cb(
    req: srpc.p.AddActionPointJointsUsingRobot.Request, ui: WsClient
) -> None:

    hlp.is_valid_identifier(req.args.name)
    proj = glob.LOCK.project_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):

        ensure_scene_started()

        robot_inst = get_robot_instance(req.args.robot_id)
        await check_eef_arm(robot_inst, req.args.arm_id)

        ap = proj.bare_action_point(req.args.action_point_id)

        unique_name(req.args.name, proj.ap_joint_names(ap.id))

        new_joints = await get_robot_joints(robot_inst, req.args.arm_id)

        await ensure_write_locked(ap.id, glob.USERS.user_name(ui))

        if req.dry_run:
            return None

        prj = common.ProjectRobotJoints(
            req.args.name, req.args.robot_id, new_joints, True, req.args.arm_id, req.args.end_effector_id
        )
        proj.upsert_joints(ap.id, prj)

        evt = sevts.p.JointsChanged(prj)
        evt.change_type = Event.Type.ADD
        evt.parent_id = ap.id
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_action_point_joints_using_robot_cb(
    req: srpc.p.UpdateActionPointJointsUsingRobot.Request, ui: WsClient
) -> None:

    proj = glob.LOCK.project_or_exception()

    ensure_scene_started()

    ap, robot_joints = proj.ap_and_joints(req.args.joints_id)

    user_name = glob.USERS.user_name(ui)

    async with ctx_read_lock(robot_joints.robot_id, user_name):
        await ensure_write_locked(ap.id, user_name)

        robot_joints.joints = await get_robot_joints(get_robot_instance(robot_joints.robot_id), robot_joints.arm_id)
        robot_joints.is_valid = True

        proj.update_modified()

        evt = sevts.p.JointsChanged(robot_joints)
        evt.change_type = Event.Type.UPDATE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_action_point_joints_cb(req: srpc.p.UpdateActionPointJoints.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    ap, robot_joints = proj.ap_and_joints(req.args.joints_id)

    if {joint.name for joint in req.args.joints} != {joint.name for joint in robot_joints.joints}:
        raise Arcor2Exception("Joint names does not match the robot.")

    await ensure_write_locked(ap.id, glob.USERS.user_name(ui))

    # TODO maybe joints values should be normalized? To <0, 2pi> or to <-pi, pi>?
    robot_joints.joints = req.args.joints
    robot_joints.is_valid = True
    proj.update_modified()

    evt = sevts.p.JointsChanged(robot_joints)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def remove_action_point_joints_cb(req: srpc.p.RemoveActionPointJoints.Request, ui: WsClient) -> None:
    """Removes joints from action point.

    :param req:
    :return:
    """

    proj = glob.LOCK.project_or_exception()

    for act in proj.actions:
        for param in act.parameters:
            if plugin_from_type_name(param.type).uses_robot_joints(proj, act.id, param.name, req.args.joints_id):
                raise Arcor2Exception(f"Joints used in action {act.name} (parameter {param.name}).")

    ap, _ = proj.ap_and_joints(req.args.joints_id)
    await ensure_write_locked(ap.id, glob.USERS.user_name(ui))

    joints_to_be_removed = proj.remove_joints(req.args.joints_id)

    proj.update_modified()

    evt = sevts.p.JointsChanged(joints_to_be_removed)
    evt.change_type = Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def rename_action_point_cb(req: srpc.p.RenameActionPoint.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    ap = proj.bare_action_point(req.args.action_point_id)

    if req.args.new_name == ap.name:
        raise Arcor2Exception("Name unchanged")

    hlp.is_valid_identifier(req.args.new_name)
    unique_name(req.args.new_name, proj.action_points_names)

    user_name = glob.USERS.user_name(ui)
    await ensure_write_locked(req.args.action_point_id, user_name)

    if req.dry_run:
        return None

    ap.name = req.args.new_name

    proj.update_modified()

    await glob.LOCK.write_unlock(req.args.action_point_id, user_name, True)

    evt = sevts.p.ActionPointChanged(ap)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None


def detect_ap_loop(proj: CachedProject, ap: common.BareActionPoint, new_parent_id: str) -> None:

    visited_ids: set[str] = set()
    ap = copy.deepcopy(ap)
    ap.parent = new_parent_id
    while True:

        if ap.id in visited_ids:
            raise Arcor2Exception("Loop detected!")

        visited_ids.add(ap.id)

        if ap.parent is None:
            break  # type: ignore  # this is certainly reachable

        try:
            ap = proj.bare_action_point(ap.parent)
        except Arcor2Exception:
            break


async def update_action_point_parent_cb(req: srpc.p.UpdateActionPointParent.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)
    ap = proj.bare_action_point(req.args.action_point_id)

    # some super basic checks (not involving other objects) are performed before locking anything
    if req.args.new_parent_id == ap.parent:
        return None

    if req.args.new_parent_id == ap.id:
        raise Arcor2Exception("AP can't be its own parent.")

    to_lock = await get_unlocked_objects(ap.parent, user_name) if ap.parent else set()
    if req.args.new_parent_id:
        to_lock.add(req.args.new_parent_id)

    async with ctx_write_lock(to_lock, user_name):

        check_ap_parent(scene, proj, req.args.new_parent_id)
        detect_ap_loop(proj, ap, req.args.new_parent_id)

        await ensure_write_locked(ap.id, user_name)  # TODO should require locked tree?

        if req.dry_run:
            return

        # Save root and parent of current AP and apply it to structures after successful update
        current_root = await glob.LOCK.get_root_id(ap.id)
        old_parent = ap.parent

        if not ap.parent and req.args.new_parent_id:
            # AP position and all orientations will become relative to the parent
            updated_aps = tr.make_global_ap_relative(scene, proj, ap, req.args.new_parent_id)

        elif ap.parent and not req.args.new_parent_id:
            # AP position and all orientations will become absolute
            updated_aps = tr.make_relative_ap_global(scene, proj, ap)
        else:

            assert ap.parent
            # AP position and all orientations will become relative to another parent
            _updated_aps = tr.make_relative_ap_global(scene, proj, ap)
            updated_aps = tr.make_global_ap_relative(scene, proj, ap, req.args.new_parent_id)

            updated_aps = set.union(updated_aps, _updated_aps)

        proj.update_child(ap.id, old_parent, req.args.new_parent_id)
        await glob.LOCK.update_write_lock(ap.id, current_root, user_name)

        assert (req.args.new_parent_id and ap.parent == req.args.new_parent_id) or (
            req.args.new_parent_id == "" and ap.parent is None
        ), f"Requested parent id: {req.args.new_parent_id}, current parent id: {ap.parent}"
        proj.update_modified()

        """
        Can't send orientation changes and then ActionPointChanged/UPDATE_BASE (or vice versa)
        because UI would display orientations wrongly (for a short moment).
        """
        # 'ap' is BareActionPoint, that does not contain orientations
        evt = sevts.p.ActionPointChanged(proj.action_point(req.args.action_point_id))
        evt.change_type = Event.Type.UPDATE
        asyncio.ensure_future(notif.broadcast_event(evt))

        for cap in updated_aps:
            evt = sevts.p.ActionPointChanged(proj.action_point(cap))
            evt.change_type = Event.Type.UPDATE
            asyncio.ensure_future(notif.broadcast_event(evt))

    await glob.LOCK.write_unlock(ap.id, user_name, True)

    return None


async def update_ap_position(
    proj: UpdateableCachedProject, ap: common.BareActionPoint, position: common.Position
) -> None:
    """Updates position of an AP and sends notification about joints that
    become invalid because of it.

    :param ap:
    :param position:
    :return:
    """

    valid_joints = [joints for joints in proj.ap_joints(ap.id) if joints.is_valid]
    proj.update_ap_position(ap.id, position)

    for joints in valid_joints:  # those are now invalid, so let's notify UI about the change

        assert not joints.is_valid

        evt = sevts.p.JointsChanged(joints)
        evt.change_type = Event.Type.UPDATE
        evt.parent_id = ap.id
        asyncio.ensure_future(notif.broadcast_event(evt))

    ap_evt = sevts.p.ActionPointChanged(ap)
    ap_evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(ap_evt))


async def update_action_point_position_cb(req: srpc.p.UpdateActionPointPosition.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    ap = proj.bare_action_point(req.args.action_point_id)
    user_name = glob.USERS.user_name(ui)

    if req.dry_run:
        await glob.LOCK.check_lock_tree(req.args.action_point_id, user_name)
    else:
        await ensure_write_locked(req.args.action_point_id, user_name, True)

    if req.dry_run:
        return

    await update_ap_position(proj, ap, req.args.new_position)


async def update_action_point_using_robot_cb(req: srpc.p.UpdateActionPointUsingRobot.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()
    ensure_scene_started()
    user_name = glob.USERS.user_name(ui)

    ap = proj.bare_action_point(req.args.action_point_id)
    await ensure_write_locked(ap.id, user_name)

    async with ctx_read_lock(req.args.robot.robot_id, user_name):
        robot_inst = get_robot_instance(req.args.robot.robot_id)
        await check_eef_arm(robot_inst, req.args.robot.arm_id, req.args.robot.end_effector)
        new_pose = await get_end_effector_pose(robot_inst, req.args.robot.end_effector, req.args.robot.arm_id)

        if ap.parent:
            new_pose = tr.make_pose_rel_to_parent(scene, proj, new_pose, ap.parent)

        await update_ap_position(proj, proj.bare_action_point(req.args.action_point_id), new_pose.position)
        return None


async def add_action_point_orientation_cb(req: srpc.p.AddActionPointOrientation.Request, ui: WsClient) -> None:
    """Adds orientation and joints to the action point.

    :param req:
    :return:
    """

    proj = glob.LOCK.project_or_exception()

    ap = proj.bare_action_point(req.args.action_point_id)
    hlp.is_valid_identifier(req.args.name)
    unique_name(req.args.name, proj.ap_orientation_names(ap.id))

    await ensure_write_locked(req.args.action_point_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    orientation = common.NamedOrientation(req.args.name, req.args.orientation)
    proj.upsert_orientation(ap.id, orientation)

    evt = sevts.p.OrientationChanged(orientation)
    evt.change_type = Event.Type.ADD
    evt.parent_id = ap.id
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def update_action_point_orientation_cb(req: srpc.p.UpdateActionPointOrientation.Request, ui: WsClient) -> None:
    """Updates orientation of the action point.

    :param req:
    :return:
    """

    proj = glob.LOCK.project_or_exception()
    orientation = proj.orientation(req.args.orientation_id)
    await ensure_write_locked(orientation.id, glob.USERS.user_name(ui))

    orientation.orientation = req.args.orientation

    proj.update_modified()

    evt = sevts.p.OrientationChanged(orientation)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def add_action_point_orientation_using_robot_cb(
    req: srpc.p.AddActionPointOrientationUsingRobot.Request, ui: WsClient
) -> None:
    """Adds orientation and joints to the action point.

    :param req:
    :return:
    """

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)

    async with ctx_read_lock(req.args.robot.robot_id, user_name):

        ensure_scene_started()

        ap = proj.bare_action_point(req.args.action_point_id)
        hlp.is_valid_identifier(req.args.name)
        unique_name(req.args.name, proj.ap_orientation_names(ap.id))
        robot_inst = get_robot_instance(req.args.robot.robot_id)
        await check_eef_arm(robot_inst, req.args.robot.arm_id, req.args.robot.end_effector)

        await ensure_write_locked(req.args.action_point_id, user_name)

        if req.dry_run:
            return None

        new_pose = await get_end_effector_pose(robot_inst, req.args.robot.end_effector, req.args.robot.arm_id)

        if ap.parent:
            new_pose = tr.make_pose_rel_to_parent(scene, proj, new_pose, ap.parent)

        orientation = common.NamedOrientation(req.args.name, new_pose.orientation)
        proj.upsert_orientation(ap.id, orientation)

        evt = sevts.p.OrientationChanged(orientation)
        evt.change_type = Event.Type.ADD
        evt.parent_id = ap.id
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_action_point_orientation_using_robot_cb(
    req: srpc.p.UpdateActionPointOrientationUsingRobot.Request, ui: WsClient
) -> None:
    """Updates orientation and joint of the action point.

    :param req:
    :return:
    """

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)

    async with ctx_read_lock(req.args.robot.robot_id, user_name):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot.robot_id)
        await check_eef_arm(robot_inst, req.args.robot.arm_id, req.args.robot.end_effector)

        ap, ori = proj.bare_ap_and_orientation(req.args.orientation_id)

        await ensure_write_locked(ori.id, user_name)

        new_pose = await get_end_effector_pose(robot_inst, req.args.robot.end_effector, req.args.robot.arm_id)

        if ap.parent:
            new_pose = tr.make_pose_rel_to_parent(scene, proj, new_pose, ap.parent)

        ori.orientation = new_pose.orientation

        proj.update_modified()

        evt = sevts.p.OrientationChanged(ori)
        evt.change_type = Event.Type.UPDATE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def remove_action_point_orientation_cb(req: srpc.p.RemoveActionPointOrientation.Request, ui: WsClient) -> None:
    """Removes orientation.

    :param req:
    :return:
    """

    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)
    ap, orientation = proj.bare_ap_and_orientation(req.args.orientation_id)

    to_lock = await get_unlocked_objects(orientation.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):
        for act in proj.actions:
            for param in act.parameters:
                if plugin_from_type_name(param.type).uses_orientation(
                    proj, act.id, param.name, req.args.orientation_id
                ):
                    raise Arcor2Exception(f"Orientation used in action {act.name} (parameter {param.name}).")

        if req.dry_run:
            return None

        await glob.LOCK.write_unlock(orientation.id, user_name)

        proj.remove_orientation(req.args.orientation_id)

        evt = sevts.p.OrientationChanged(orientation)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def open_project_cb(req: srpc.p.OpenProject.Request, ui: WsClient) -> None:
    """Opens a project. This can be done even when a scene or another project
    is opened.

    :param req:
    :param ui:
    :return:
    """

    async with glob.LOCK.get_lock():

        if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
            raise Arcor2Exception("Can't open project while package runs.")

        await open_project(req.args.id)

        assert glob.LOCK.scene
        assert glob.LOCK.project

        asyncio.ensure_future(
            notify_project_opened(
                sevts.p.OpenProject(sevts.p.OpenProject.Data(glob.LOCK.scene.scene, glob.LOCK.project.project))
            )
        )

        return None


async def save_project_cb(req: srpc.p.SaveProject.Request, ui: WsClient) -> None:

    async with glob.LOCK.get_lock(req.dry_run):

        proj = glob.LOCK.project_or_exception()

        if proj.modified and not proj.has_changes:
            raise Arcor2Exception("No changes to save.")

        if req.dry_run:
            return None

        # TODO temporary code to help debugging long lasting update of project
        import time

        start = time.monotonic()
        proj.modified = await storage.update_project(proj)
        logger.info(f"Updating the project took {time.monotonic()-start:.3f}s.")

    asyncio.ensure_future(notif.broadcast_event(sevts.p.ProjectSaved()))
    return None


async def new_project_cb(req: srpc.p.NewProject.Request, ui: WsClient) -> None:

    if glob.LOCK.project:
        raise Arcor2Exception("Project has to be closed first.")

    async with ctx_write_lock(glob.LOCK.SpecialValues.PROJECT, glob.USERS.user_name(ui), dry_run=req.dry_run):
        if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
            raise Arcor2Exception("Can't create project while package runs.")

        unique_name(req.args.name, (await project_names()))

        # TODO workaround for https://gitlab.com/kinalisoft/test-it-off/project/-/issues/16
        unique_name(req.args.name, (await scene_names()))

        if await glob.LOCK.get_write_locks_count() > 1:  # project lock also counts
            raise Arcor2Exception("Project has locked objects")

        if req.dry_run:
            return None

        if glob.LOCK.scene:
            if glob.LOCK.scene.id != req.args.scene_id:
                raise Arcor2Exception("Another scene is opened.")

            # the scene might not be saved yet - another check
            # TODO workaround for https://gitlab.com/kinalisoft/test-it-off/project/-/issues/16
            if req.args.name == glob.LOCK.scene.name:
                raise Arcor2Exception("A project can't have the same name as the scene.")

            if glob.LOCK.scene.has_changes:
                await save_scene()
        else:

            if req.args.scene_id not in (await storage.get_scene_ids()):
                raise Arcor2Exception("Unknown scene id.")

            await open_scene(req.args.scene_id)

        glob.PREV_RESULTS.clear()
        glob.LOCK.project = UpdateableCachedProject(
            common.Project(
                req.args.name, req.args.scene_id, description=req.args.description, has_logic=req.args.has_logic
            )
        )

        assert glob.LOCK.scene

        # add commonly used project parameters
        from arcor2 import json  # TODO use parameter plugin

        glob.LOCK.project.upsert_parameter(
            common.ProjectParameter("scene_id", "string", json.dumps(glob.LOCK.scene.id))
        )
        glob.LOCK.project.upsert_parameter(
            common.ProjectParameter("project_id", "string", json.dumps(glob.LOCK.project.id))
        )

        asyncio.ensure_future(
            notify_project_opened(
                sevts.p.OpenProject(sevts.p.OpenProject.Data(glob.LOCK.scene.scene, glob.LOCK.project.project))
            )
        )
        return None


async def close_project_cb(req: srpc.p.CloseProject.Request, ui: WsClient) -> None:

    async with glob.LOCK.get_lock(req.dry_run):

        proj = glob.LOCK.project_or_exception()
        can_modify_scene()

        if not req.args.force and proj.has_changes:
            raise Arcor2Exception("Project has unsaved changes.")

        if req.dry_run:
            return None

        await close_project()
        return None


async def add_action_point_cb(req: srpc.p.AddActionPoint.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    async with ctx_write_lock(glob.LOCK.SpecialValues.PROJECT, glob.USERS.user_name(ui)):

        hlp.is_valid_identifier(req.args.name)
        unique_name(req.args.name, proj.action_points_names)
        check_ap_parent(scene, proj, req.args.parent)

        if req.dry_run:
            return None

        ap = proj.upsert_action_point(common.ActionPoint.uid(), req.args.name, req.args.position, req.args.parent)

        evt = sevts.p.ActionPointChanged(ap)
        evt.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def remove_action_point_cb(req: srpc.p.RemoveActionPoint.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects(req.args.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):

        ap = proj.bare_action_point(req.args.id)

        for proj_ap in proj.action_points_with_parent:
            if proj_ap.parent == ap.id:
                raise Arcor2Exception(f"Can't remove parent of '{proj_ap.name}' AP.")

        ap_action_ids = proj.ap_action_ids(ap.id)

        # check if AP's actions aren't involved in logic
        # TODO 'force' param to remove logical connections?
        for logic in proj.logic:
            if (
                logic.start in ap_action_ids
                or logic.end in ap_action_ids
                or (logic.condition and logic.condition.parse_what().action_id in ap_action_ids)
            ):
                raise Arcor2Exception("Remove logic connections first.")

        for act in proj.actions:

            if act.id in ap_action_ids:
                continue

            for param in act.parameters:
                if param.type == common.ActionParameter.TypeEnum.LINK:
                    parsed_link = param.parse_link()
                    linking_action = proj.action(parsed_link.action_id)
                    if parsed_link.action_id in ap_action_ids:
                        raise Arcor2Exception(f"Result of '{act.name}' is linked from '{linking_action.name}'.")

                if not param.is_value():
                    continue

                for joints in proj.ap_joints(ap.id):
                    if plugin_from_type_name(param.type).uses_robot_joints(proj, act.id, param.name, joints.id):
                        raise Arcor2Exception(
                            f"Joints {joints.name} used in action {act.name} (parameter {param.name})."
                        )

                for ori in proj.ap_orientations(ap.id):
                    if plugin_from_type_name(param.type).uses_orientation(proj, act.id, param.name, ori.id):
                        raise Arcor2Exception(
                            f"Orientation {ori.name} used in action {act.name} (parameter {param.name})."
                        )

                # TODO some hypothetical parameter type could use just bare ActionPoint (its position)

        if not await glob.LOCK.check_remove(ap.id, user_name):
            raise Arcor2Exception("Children locked")

        if req.dry_run:
            return None

        await glob.LOCK.write_unlock(req.args.id, user_name)

        proj.remove_action_point(req.args.id)

        evt = sevts.p.ActionPointChanged(ap)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def copy_action_point_cb(req: srpc.p.CopyActionPoint.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    async def copy_action_point(
        orig_ap: common.BareActionPoint,
        new_parent_id: None | str = None,
        position: None | common.Position = None,
    ) -> None:

        assert await glob.LOCK.is_read_locked(orig_ap.id, user_name)

        ap = proj.upsert_action_point(
            common.ActionPoint.uid(),
            make_name_unique(f"{orig_ap.name}_copy", proj.action_points_names),
            orig_ap.position if position is None else position,
            orig_ap.parent if new_parent_id is None else new_parent_id,
        )

        ap_added_evt = sevts.p.ActionPointChanged(ap)
        ap_added_evt.change_type = Event.Type.ADD
        await notif.broadcast_event(ap_added_evt)

        for ori in proj.ap_orientations(orig_ap.id):

            assert await glob.LOCK.is_read_locked(ori.id, user_name)

            new_ori = ori.copy()
            old_ori_to_new_ori[ori.id] = new_ori.id
            proj.upsert_orientation(ap.id, new_ori)

            ori_added_evt = sevts.p.OrientationChanged(new_ori)
            ori_added_evt.change_type = Event.Type.ADD
            ori_added_evt.parent_id = ap.id
            await notif.broadcast_event(ori_added_evt)

        for joints in proj.ap_joints(orig_ap.id):

            assert await glob.LOCK.is_read_locked(joints.id, user_name)

            new_joints = joints.copy()
            proj.upsert_joints(ap.id, new_joints)

            joints_added_evt = sevts.p.JointsChanged(new_joints)
            joints_added_evt.change_type = Event.Type.ADD
            joints_added_evt.parent_id = ap.id
            await notif.broadcast_event(joints_added_evt)

        action_names = proj.action_names  # action name has to be globally unique
        for act in proj.ap_actions(orig_ap.id):
            assert await glob.LOCK.is_read_locked(act.id, user_name)
            new_act = act.copy()
            new_act.name = make_name_unique(f"{act.name}_copy", action_names)
            proj.upsert_action(ap.id, new_act)
            new_action_ids.add(new_act.id)

        # find direct child-APs and copy them too
        # this can't be done in parallel for all child APs (whole hierarchy), because UI would be confused
        await asyncio.gather(
            *[
                copy_action_point(proj.action_point(child_id), ap.id)
                for child_id in proj.childs(orig_ap.id)
                if child_id in proj.action_points_ids
            ]
        )

    user_name = glob.USERS.user_name(ui)
    childs_of_original_ap = glob.LOCK.get_all_children(req.args.id)

    async with ctx_read_lock(childs_of_original_ap | {req.args.id}, user_name):

        # filter out only APs
        child_aps = {ch for ch in childs_of_original_ap if ch in proj.action_points_ids} | {req.args.id}
        logger.debug(f"Child action points of the original AP: {child_aps}")

        original_ap = proj.bare_action_point(req.args.id)

        if req.dry_run:
            return

        old_ori_to_new_ori: dict[str, str] = {}
        new_action_ids: set[str] = set()
        # let's do it within the RPC, should not take long time...
        await copy_action_point(original_ap, position=req.args.position)

        for act_id in new_action_ids:  # this has to be done once old_ori_to_new_ori is complete

            ap, new_act = proj.action_point_and_action(act_id)

            for param in new_act.parameters:

                if param.type != PosePlugin.type_name():  # TODO also handle joints and positions!
                    continue

                old_ori = proj.bare_ap_and_orientation(PosePlugin.orientation_id(proj, new_act.id, param.name))[1]
                child_orientations = {ori.id for ap_id in child_aps for ori in proj.ap_orientations(ap_id)}
                logger.debug(f"Child orientations of the original AP: {child_orientations}")

                if old_ori.id not in child_orientations:
                    logger.debug(
                        f"Orientation {old_ori.name} ({old_ori.id}) "
                        f"belongs to another AP (sub)tree, no need to update anything."
                    )
                    continue

                # orientation belongs to the tree being copied so it has to be updated

                # TODO this is hacky - plugins are missing methods to set/update parameters
                from arcor2 import json

                try:
                    new_ori_id = old_ori_to_new_ori[old_ori.id]
                except KeyError:
                    logger.error(f"Failed to find a new orientation ID for {old_ori.id}.")
                else:
                    logger.debug(f"Updating orientation from {old_ori.id} to {new_ori_id}.")
                    param.value = json.dumps(new_ori_id)

            action_added_evt = sevts.p.ActionChanged(new_act)
            action_added_evt.change_type = Event.Type.ADD
            action_added_evt.parent_id = ap.id
            await notif.broadcast_event(action_added_evt)


async def add_action_cb(req: srpc.p.AddAction.Request, ui: WsClient) -> None:
    """Adds new action to project.

    Used also when duplicating action.
    """

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    # When duplicating action AP cannot be removed, no need to lock
    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects(req.args.action_point_id, user_name)
    async with ctx_write_lock(to_lock, user_name):
        ap = proj.bare_action_point(req.args.action_point_id)

        unique_name(req.args.name, proj.action_names)

        new_action = common.Action(req.args.name, req.args.type, parameters=req.args.parameters, flows=req.args.flows)

        action_meta = find_object_action(glob.OBJECT_TYPES, scene, new_action)

        updated_project = copy.deepcopy(proj)
        updated_project.upsert_action(req.args.action_point_id, new_action)

        check_flows(updated_project, new_action, action_meta)
        check_action_params(glob.OBJECT_TYPES, scene, updated_project, new_action, action_meta)

        if req.dry_run:
            return None

        proj.upsert_action(ap.id, new_action)

        evt = sevts.p.ActionChanged(new_action)
        evt.change_type = Event.Type.ADD
        evt.parent_id = ap.id
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_action_cb(req: srpc.p.UpdateAction.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    updated_project = copy.deepcopy(proj)

    updated_action = updated_project.action(req.args.action_id)

    if req.args.parameters is not None:
        updated_action.parameters = req.args.parameters
    if req.args.flows is not None:
        updated_action.flows = req.args.flows

    updated_action_meta = find_object_action(glob.OBJECT_TYPES, scene, updated_action)

    check_flows(updated_project, updated_action, updated_action_meta)
    check_action_params(glob.OBJECT_TYPES, scene, updated_project, updated_action, updated_action_meta)

    await ensure_write_locked(req.args.action_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    orig_action = proj.action(req.args.action_id)
    orig_action.parameters = updated_action.parameters
    proj.update_modified()

    evt = sevts.p.ActionChanged(updated_action)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def remove_action_cb(req: srpc.p.RemoveAction.Request, ui: WsClient) -> None:
    def check_action_usage(proj: CachedProject, action: common.Action) -> None:

        # check parameters
        for act in proj.actions:
            for param in act.parameters:
                if param.type == common.ActionParameter.TypeEnum.LINK:
                    link = param.parse_link()
                    if action.id == link.action_id:
                        raise Arcor2Exception(f"Action output used as parameter of {act.name}/{param.name}.")

        # check logic
        for log in proj.logic:

            if log.start == action.id or log.end == action.id:
                raise Arcor2Exception("Action used in logic.")

            if log.condition:
                action_id, _, _ = log.condition.what.split("/")

                if action_id == action.id:
                    raise Arcor2Exception("Action used in condition.")

    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects(req.args.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):

        ap, action = proj.action_point_and_action(req.args.id)
        check_action_usage(proj, action)

        if req.dry_run:
            return None

        await glob.LOCK.write_unlock(req.args.id, user_name)

        proj.remove_action(req.args.id)
        glob.remove_prev_result(action.id)

        evt = sevts.p.ActionChanged(action.bare)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def add_logic_item_cb(req: srpc.p.AddLogicItem.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception(must_have_logic=True)
    user_name = glob.USERS.user_name(ui)

    to_lock = await get_unlocked_objects([req.args.start, req.args.end], user_name)
    async with ctx_write_lock(to_lock, user_name):
        logic_item = common.LogicItem(req.args.start, req.args.end, req.args.condition)
        check_logic_item(glob.OBJECT_TYPES, scene, proj, logic_item)

        if logic_item.start != logic_item.START:
            updated_project = copy.deepcopy(proj)
            updated_project.upsert_logic_item(logic_item)
            check_for_loops(updated_project, logic_item.parse_start().start_action_id)

        if req.dry_run:
            return

        proj.upsert_logic_item(logic_item)

        evt = sevts.p.LogicItemChanged(logic_item)
        evt.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))

    await glob.LOCK.write_unlock(
        [item for item in (req.args.start, req.args.end) if item not in to_lock], user_name, True
    )
    return None


async def update_logic_item_cb(req: srpc.p.UpdateLogicItem.Request, ui: WsClient) -> None:
    # TODO lock RPC when used

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception(must_have_logic=True)

    updated_project = copy.deepcopy(proj)
    updated_logic_item = updated_project.logic_item(req.args.logic_item_id)

    updated_logic_item.start = req.args.start
    updated_logic_item.end = req.args.end
    updated_logic_item.condition = req.args.condition

    check_logic_item(glob.OBJECT_TYPES, scene, updated_project, updated_logic_item)

    if updated_logic_item.start != updated_logic_item.START:
        check_for_loops(updated_project, updated_logic_item.parse_start().start_action_id)

    if req.dry_run:
        return

    proj.upsert_logic_item(updated_logic_item)

    evt = sevts.p.LogicItemChanged(updated_logic_item)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def remove_logic_item_cb(req: srpc.p.RemoveLogicItem.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception(must_have_logic=True)
    logic_item = proj.logic_item(req.args.logic_item_id)

    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects([logic_item.start, logic_item.end], user_name)
    async with ctx_write_lock(to_lock, user_name):
        # TODO is it necessary to check something here?
        proj.remove_logic_item(req.args.logic_item_id)

        evt = sevts.p.LogicItemChanged(logic_item)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))

    await glob.LOCK.write_unlock(
        [item for item in (logic_item.start, logic_item.end) if item not in to_lock], user_name, True
    )
    return None


async def add_project_parameter_cb(req: srpc.p.AddProjectParameter.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    async with ctx_write_lock(glob.LOCK.SpecialValues.PROJECT, glob.USERS.user_name(ui)):

        pparam = common.ProjectParameter(req.args.name, req.args.type, req.args.value)
        check_project_parameter(proj, pparam)

        if req.dry_run:
            return

        proj.upsert_parameter(pparam)

        evt = sevts.p.ProjectParameterChanged(pparam)
        evt.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_project_parameter_cb(req: srpc.p.UpdateProjectParameter.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    param = proj.parameter(req.args.id)
    user_name = glob.USERS.user_name(ui)

    await ensure_write_locked(req.args.id, user_name)

    updated_param = copy.deepcopy(param)

    if req.args.name is not None:
        updated_param.name = req.args.name
    if req.args.value is not None:
        updated_param.value = req.args.value

    check_project_parameter(proj, updated_param)

    if req.dry_run:
        return

    proj.upsert_parameter(updated_param)

    evt = sevts.p.ProjectParameterChanged(updated_param)
    evt.change_type = Event.Type.UPDATE
    asyncio.create_task(notif.broadcast_event(evt))

    await glob.LOCK.write_unlock(param.id, user_name, True)


async def remove_project_parameter_cb(req: srpc.p.RemoveProjectParameter.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    pparam = proj.parameter(req.args.id)

    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects(req.args.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):

        # check for usage
        for act in proj.actions:
            for param in act.parameters:
                if (
                    param.type == common.ActionParameter.TypeEnum.PROJECT_PARAMETER
                    and param.str_from_value() == pparam.id
                ):
                    raise Arcor2Exception(f"Project parameter used in {act.name} action.")

        if req.dry_run:
            return

        await glob.LOCK.write_unlock(req.args.id, user_name)

        proj.remove_parameter(pparam.id)

        evt = sevts.p.ProjectParameterChanged(pparam)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))


async def delete_project_cb(req: srpc.p.DeleteProject.Request, ui: WsClient) -> None:

    if glob.LOCK.project:
        raise Arcor2Exception("Project has to be closed first.")

    user_name = glob.USERS.user_name(ui)

    async with ctx_write_lock(req.args.id, user_name, auto_unlock=False):
        project = await storage.get_project(req.args.id)
        await glob.LOCK.write_unlock(req.args.id, user_name)
        await storage.delete_project(req.args.id)

        evt = sevts.p.ProjectChanged(project.bare)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def rename_project_cb(req: srpc.p.RenameProject.Request, ui: WsClient) -> None:

    unique_name(req.args.new_name, (await project_names()))

    # TODO workaround for https://gitlab.com/kinalisoft/test-it-off/project/-/issues/16
    unique_name(req.args.new_name, (await scene_names()))

    user_name = glob.USERS.user_name(ui)

    await ensure_write_locked(req.args.project_id, user_name)

    if req.dry_run:
        return None

    async with managed_project(req.args.project_id) as project:

        project.name = req.args.new_name
        project.update_modified()

        evt = sevts.p.ProjectChanged(project.bare)
        evt.change_type = Event.Type.UPDATE_BASE
        asyncio.ensure_future(notif.broadcast_event(evt))

    await glob.LOCK.write_unlock(req.args.project_id, user_name)
    return None


async def copy_project_cb(req: srpc.p.CopyProject.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.source_id, glob.USERS.user_name(ui)):
        unique_name(req.args.target_name, (await project_names()))

        if req.dry_run:
            return None

        async with managed_project(req.args.source_id, make_copy=True) as project:

            project.name = req.args.target_name

            evt = sevts.p.ProjectChanged(project.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))

        return None


async def update_project_description_cb(req: srpc.p.UpdateProjectDescription.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.project_id, glob.USERS.user_name(ui)):
        async with managed_project(req.args.project_id) as project:

            project.description = req.args.new_description
            project.update_modified()

            evt = sevts.p.ProjectChanged(project.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_project_has_logic_cb(req: srpc.p.UpdateProjectHasLogic.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.project_id, glob.USERS.user_name(ui)):
        async with managed_project(req.args.project_id) as project:

            if project.has_logic and not req.args.new_has_logic:
                project.clear_logic()

                """
                TODO

                if glob.LOCK.project and glob.LOCK.project.id == req.args.project_id:
                ...send remove event for each logic item?
                """

            project.has_logic = req.args.new_has_logic
            project.update_modified()

            evt = sevts.p.ProjectChanged(project.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def rename_action_point_joints_cb(req: srpc.p.RenameActionPointJoints.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    ap, joints = proj.ap_and_joints(req.args.joints_id)

    hlp.is_valid_identifier(req.args.new_name)
    unique_name(req.args.new_name, proj.ap_joint_names(ap.id))

    await ensure_write_locked(ap.id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    joints.name = req.args.new_name
    proj.update_modified()

    evt = sevts.p.JointsChanged(joints)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None


async def rename_action_point_orientation_cb(req: srpc.p.RenameActionPointOrientation.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    ap, ori = proj.bare_ap_and_orientation(req.args.orientation_id)

    hlp.is_valid_identifier(req.args.new_name)
    unique_name(req.args.new_name, proj.ap_orientation_names(ap.id))

    await ensure_write_locked(ori.id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    ori.name = req.args.new_name
    proj.update_modified()

    evt = sevts.p.OrientationChanged(ori)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None


async def rename_action_cb(req: srpc.p.RenameAction.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    unique_name(req.args.new_name, proj.action_names)
    user_name = glob.USERS.user_name(ui)

    await ensure_write_locked(req.args.action_id, user_name)

    if req.dry_run:
        return None

    act = proj.action(req.args.action_id)
    act.name = req.args.new_name

    proj.update_modified()

    await glob.LOCK.write_unlock(req.args.action_id, user_name, True)

    evt = sevts.p.ActionChanged(act)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None



from arcor2_arserver.rpc import camera, execution, lock, objects, project, robot, scene, user

__all__ = [
    execution.__name__,
    objects.__name__,
    project.__name__,
    robot.__name__,
    scene.__name__,
    camera.__name__,
    lock.__name__,
    user.__name__,
]



import asyncio
from typing import NamedTuple

from arcor2 import helpers as hlp
from arcor2.cached import CachedScene
from arcor2.clients import aio_asset as asset
from arcor2.data.events import Event
from arcor2.data.object_type import Mesh, ObjectModel
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types import utils as otu
from arcor2.object_types.abstract import Generic, Robot
from arcor2.object_types.utils import built_in_types_names, prepare_object_types_dir
from arcor2.parameter_plugins.base import TypesDict
from arcor2.source.utils import parse
from arcor2_arserver import globals as glob
from arcor2_arserver import logger
from arcor2_arserver import notifications as notif
from arcor2_arserver import settings
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.object_types.utils import (
    ObjectTypeData,
    ObjectTypeDict,
    add_ancestor_actions,
    built_in_types_data,
    meta_from_def,
    obj_description_from_base,
    object_actions,
    remove_object_type,
)
from arcor2_arserver.robot import get_robot_meta
from arcor2_arserver_data.events.objects import ChangedObjectTypes
from arcor2_arserver_data.objects import ObjectTypeMeta


def get_types_dict() -> TypesDict:

    return {k: v.type_def for k, v in glob.OBJECT_TYPES.items() if v.type_def is not None}


def get_obj_type_data(scene: CachedScene, object_id: str) -> ObjectTypeData:
    return glob.OBJECT_TYPES[scene.object(object_id).type]


def valid_object_types() -> ObjectTypeDict:
    """To get only valid (not disabled) types.

    :return:
    """

    return {obj_type: obj for obj_type, obj in glob.OBJECT_TYPES.items() if not obj.meta.disabled}


async def get_object_data(object_types: ObjectTypeDict, obj_id: str) -> None:

    logger.debug(f"Processing {obj_id}.")

    if obj_id in object_types:
        logger.debug(f"{obj_id} already processed, skipping...")
        return

    obj_iddesc = await storage.get_object_type_iddesc(obj_id)

    if obj_id in glob.OBJECT_TYPES:

        assert obj_iddesc.modified
        assert glob.OBJECT_TYPES[obj_id].meta.modified, f"Object {obj_id} does not have 'modified' in its meta."

        if obj_iddesc.modified == glob.OBJECT_TYPES[obj_id].meta.modified:
            logger.debug(f"No need to update {obj_id}.")
            return

    obj = await storage.get_object_type(obj_id)

    try:
        bases = otu.base_from_source(obj.source, obj_id)

        if not bases:
            logger.debug(f"{obj_id} is definitely not an ObjectType (subclass of {object.__name__}), maybe mixin?")
            return

        if bases[0] not in object_types.keys() | built_in_types_names():
            logger.debug(f"Getting base class {bases[0]} for {obj_id}.")
            await get_object_data(object_types, bases[0])

        for mixin in bases[1:]:
            mixin_obj = await storage.get_object_type(mixin)

            await hlp.run_in_executor(
                hlp.save_and_import_type_def,
                mixin_obj.source,
                mixin_obj.id,
                object,
                settings.OBJECT_TYPE_PATH,
                settings.OBJECT_TYPE_MODULE,
            )

    except Arcor2Exception as e:
        logger.error(f"Disabling ObjectType {obj.id}: can't get a base. {str(e)}")
        object_types[obj_id] = ObjectTypeData(
            ObjectTypeMeta(
                obj_id, "ObjectType disabled.", disabled=True, problem="Can't get base.", modified=obj.modified
            )
        )
        return

    logger.debug(f"Updating {obj_id}.")

    try:
        type_def = await hlp.run_in_executor(
            hlp.save_and_import_type_def,
            obj.source,
            obj.id,
            Generic,
            settings.OBJECT_TYPE_PATH,
            settings.OBJECT_TYPE_MODULE,
        )
    except Arcor2Exception as e:
        logger.debug(f"{obj.id} is probably not an ObjectType. {str(e)}")
        return

    assert issubclass(type_def, Generic)

    try:
        meta = meta_from_def(type_def)
    except Arcor2Exception as e:
        logger.error(f"Disabling ObjectType {obj.id}.")
        logger.debug(e, exc_info=True)
        object_types[obj_id] = ObjectTypeData(
            ObjectTypeMeta(obj_id, "ObjectType disabled.", disabled=True, problem=str(e), modified=obj.modified)
        )
        return

    meta.modified = obj.modified

    if obj.model:
        try:
            model = await storage.get_model(obj.model.id, obj.model.type)
        except Arcor2Exception as e:
            logger.error(f"{obj.model.id}: failed to get collision model of type {obj.model.type}. {str(e)}")
            meta.disabled = True
            meta.problem = "Can't get collision model."
            object_types[obj_id] = ObjectTypeData(meta)
            return

        if isinstance(model, Mesh) and not await asset.asset_exists(model.asset_id):
            logger.error(f"Disabling {meta.type} as its mesh file {model.asset_id} does not exist.")
            meta.disabled = True
            meta.problem = "Mesh file does not exist."
            object_types[obj_id] = ObjectTypeData(meta)
            return

        kwargs = {model.type().value.lower(): model}
        meta.object_model = ObjectModel(model.type(), **kwargs)  # type: ignore

    ast = parse(obj.source)
    otd = ObjectTypeData(meta, type_def, object_actions(type_def, ast), ast)

    object_types[obj_id] = otd


class UpdatedObjectTypes(NamedTuple):

    new: set[str]
    updated: set[str]
    removed: set[str]

    @property
    def all(self) -> set[str]:
        return self.new | self.updated | self.removed


async def get_object_types() -> UpdatedObjectTypes:
    """Serves to initialize or update knowledge about awailable ObjectTypes.

    :return:
    """

    initialization = False

    # initialize with built-in types, this has to be done just once
    if not glob.OBJECT_TYPES:
        logger.debug("Initialization of ObjectTypes.")
        initialization = True
        await hlp.run_in_executor(prepare_object_types_dir, settings.OBJECT_TYPE_PATH, settings.OBJECT_TYPE_MODULE)
        glob.OBJECT_TYPES.update(built_in_types_data())

    updated_object_types: ObjectTypeDict = {}

    object_type_ids: set[str] | list[str] = await storage.get_object_type_ids()

    if __debug__:  # this should uncover potential problems with order in which ObjectTypes are processed
        import random

        object_type_ids = list(object_type_ids)
        random.shuffle(object_type_ids)

    for obj_id in object_type_ids:
        await get_object_data(updated_object_types, obj_id)

    removed_object_ids = {
        obj for obj in glob.OBJECT_TYPES.keys() if obj not in object_type_ids
    } - built_in_types_names()
    updated_object_ids = {k for k in updated_object_types.keys() if k in glob.OBJECT_TYPES}
    new_object_ids = {k for k in updated_object_types.keys() if k not in glob.OBJECT_TYPES}

    logger.debug(f"Removed ids: {removed_object_ids}")
    logger.debug(f"Updated ids: {updated_object_ids}")
    logger.debug(f"New ids: {new_object_ids}")

    if not initialization and removed_object_ids:

        # TODO remove it from sys.modules

        remove_evt = ChangedObjectTypes([v.meta for k, v in glob.OBJECT_TYPES.items() if k in removed_object_ids])
        remove_evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(remove_evt))

        for removed in removed_object_ids:
            assert removed not in built_in_types_names(), "Attempt to remove built-in type."
            del glob.OBJECT_TYPES[removed]
            await remove_object_type(removed)

    glob.OBJECT_TYPES.update(updated_object_types)

    logger.debug(f"All known ids: {glob.OBJECT_TYPES.keys()}")

    for obj_type in updated_object_types.values():

        # if description is missing, try to get it from ancestor(s)
        if not obj_type.meta.description:

            try:
                obj_type.meta.description = obj_description_from_base(glob.OBJECT_TYPES, obj_type.meta)
            except otu.DataError as e:
                logger.error(f"Failed to get info from base for {obj_type}, error: '{e}'.")

        if not obj_type.meta.disabled and not obj_type.meta.built_in:
            add_ancestor_actions(obj_type.meta.type, glob.OBJECT_TYPES)

    if not initialization:

        if updated_object_ids:
            update_evt = ChangedObjectTypes([v.meta for k, v in glob.OBJECT_TYPES.items() if k in updated_object_ids])
            update_evt.change_type = Event.Type.UPDATE
            asyncio.ensure_future(notif.broadcast_event(update_evt))

        if new_object_ids:
            add_evt = ChangedObjectTypes([v.meta for k, v in glob.OBJECT_TYPES.items() if k in new_object_ids])
            add_evt.change_type = Event.Type.ADD
            asyncio.ensure_future(notif.broadcast_event(add_evt))

    for obj_type in updated_object_types.values():

        if obj_type.type_def and issubclass(obj_type.type_def, Robot) and not obj_type.type_def.abstract():
            await get_robot_meta(obj_type)

    # if object does not change but its base has changed, it has to be reloaded
    for obj_id, obj in glob.OBJECT_TYPES.items():

        if obj_id in updated_object_ids:
            continue

        if obj.type_def and obj.meta.base in updated_object_ids:

            logger.debug(f"Re-importing {obj.meta.type} because its base {obj.meta.base} type has changed.")
            obj.type_def = await hlp.run_in_executor(
                hlp.import_type_def,
                obj.meta.type,
                Generic,
                settings.OBJECT_TYPE_PATH,
                settings.OBJECT_TYPE_MODULE,
            )

    return UpdatedObjectTypes(new_object_ids, updated_object_ids, removed_object_ids)


async def update_object_model(meta: ObjectTypeMeta, om: ObjectModel) -> None:

    if not meta.object_model:
        return

    model = om.model()

    if model.id != meta.type:
        raise Arcor2Exception("Model id must be equal to ObjectType id.")

    if isinstance(model, Mesh):
        if not await asset.asset_exists(model.asset_id):
            raise Arcor2Exception(f"File {model.asset_id} associated to mesh {model.id} does not exist.")

    # when updating model of an already existing object, the type might be different
    if meta.object_model.type != om.type:
        await storage.delete_model(model.id)  # ...otherwise it is going to be an orphan

    await storage.put_model(model)



from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver_data import events as sevts


async def run_lock_notification_worker() -> None:
    while True:
        notif_data = await glob.LOCK.notifications_q.get()
        obj_ids = notif_data.obj_ids
        data = sevts.lk.LockData(obj_ids, notif_data.owner)

        evt = sevts.lk.ObjectsLocked(data) if notif_data.lock else sevts.lk.ObjectsUnlocked(data)

        if notif_data.client:
            await notif.event(notif_data.client, evt)
        else:
            await notif.broadcast_event(evt)



import inspect
from enum import Enum

from apispec import APISpec
from apispec.exceptions import DuplicateComponentNameError
from apispec_webframeworks.flask import FlaskPlugin
from dataclasses_jsonschema import JsonSchemaMixin
from dataclasses_jsonschema.apispec import DataclassesPlugin

from arcor2.data.events import Event
from arcor2.data.rpc.common import RPC


def _rename_childs(obj: type[JsonSchemaMixin] | type[Enum]) -> None:

    for _, child_obj in inspect.getmembers(obj, inspect.isclass):

        if issubclass(child_obj, (JsonSchemaMixin, Enum)):

            if hasattr(child_obj, "__renamed__"):
                continue

            if "." in child_obj.__qualname__:
                child_obj.__name__ = "".join(child_obj.__qualname__.split("."))
                child_obj.__renamed__ = None

            _rename_childs(child_obj)


def _add_to_spec(spec, obj) -> None:

    try:
        spec.components.schema(obj.__name__, schema=obj)
    except DuplicateComponentNameError:
        pass


def generate_openapi(service_name: str, version: str, rpcs: list[type[RPC]], events: list[type[Event]]) -> str:
    """Generate OpenAPI models from RPCs and events.

    Be aware: it modifies __name__ attribute!
    """

    # Create an APISpec
    spec = APISpec(
        title=f"{service_name} Data Models",
        version=version,
        openapi_version="3.0.2",
        plugins=[FlaskPlugin(), DataclassesPlugin()],
    )

    for obj in events:

        if obj is Event:
            continue

        _rename_childs(obj)

    for rpc in rpcs:

        if rpc is RPC:
            continue

        for cls in (rpc.Request, rpc.Response):

            cls.__name__ = rpc.__name__ + cls.__name__
            _rename_childs(cls)

    for obj in events:
        _add_to_spec(spec, obj)
    for rpc in rpcs:
        for cls in (rpc.Request, rpc.Response):
            _add_to_spec(spec, cls)

    return spec.to_yaml()



import asyncio

from websockets.server import WebSocketServerProtocol

from arcor2 import ws_server
from arcor2.data import events
from arcor2_arserver import globals as glob
from arcor2_arserver import logger


async def broadcast_event(event: events.Event) -> None:

    logger.debug(event)

    if glob.USERS.interfaces:
        message = event.to_json()
        await asyncio.gather(*[ws_server.send_json_to_client(intf, message) for intf in glob.USERS.interfaces])


async def event(interface: WebSocketServerProtocol, event: events.Event) -> None:
    await ws_server.send_json_to_client(interface, event.to_json())
